/*
 * SimulatorGUI class: AperiodicWaveletAverage
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import java.util.Observable;
import java.util.StringTokenizer;

/*
 * this is an object that will compute the apreiodic average and store the
 * appropriate value for retrieval
 */
public class AperiodicWaveletAverage extends AverageObject {

  // uniform grid on the unit cubic constructor for the class
  public AperiodicWaveletAverage(Double[][] d, int nx, int ny, int nz, int t)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         null, null, null);
  }

  // uniform grid on a specified region constructor for the class
  public AperiodicWaveletAverage(Double[][] d, int nx, int ny, int nz, int t,
                           Double xl, Double xr,
                           Double yl, Double yr,
                           Double zl, Double zr)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         xl, xr, yl, yr, zl, zr,
         null, null, null);
  }

  // specified grid constructor for the class
  public AperiodicWaveletAverage(Double[][] d, int nx, int ny, int nz, int t,
                           Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         xg, yg, zg);
  }

  // full constructor for the class
  public AperiodicWaveletAverage(Double[][] d, int nx, int ny, int nz, int t,
                           Double xl, Double xr,
                           Double yl, Double yr,
                           Double zl, Double zr,
                           Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    data = d;
    nxhet = nx;
    nyhet = ny;
    nzhet = nz;
    tag = t;
    nxnyhet = nxhet * nyhet;
    nnhet = nxnyhet * nzhet;
    if(data.length != nnhet)
     throw
       new BadArraySizeException("Array wrong size in AperiodicWaveletAverage");

    // set up the grid
    if(xl != null) {
      xleft = xl;
    } else {
      xleft = new Double(0.0);
    }
    if(xr != null) {
      xright= xr;
    } else {
      xright = new Double(1.0);
    }
    if(yl != null) {
      yleft = yl;
    } else {
      yleft = new Double(0.0);
    }
    if(yr != null) {
      yright= yr;
    } else {
      yright = new Double(1.0);
    }
    if(zl != null) {
      zleft = zl;
    } else {
      zleft = new Double(0.0);
    }
    if(zr != null) {
      zright= zr;
    } else {
      zright = new Double(1.0);
    }

    // make sure the end points are ok
    if(xleft.doubleValue() > xright.doubleValue()) {
      Double temp = xleft;
      xleft = xright;
      xright = temp;
    }
    if(yleft.doubleValue() > yright.doubleValue()) {
      Double temp = yleft;
      yleft = yright;
      yright = temp;
    }
    if(zleft.doubleValue() > zright.doubleValue()) {
      Double temp = zleft;
      zleft = zright;
      zright = temp;
    }

    // set or compute the grid lines in the domain
    if(xg != null) {
      xgrid = xg;
      nxhet = xgrid.length;
    } else {
      xgrid = computeGrid(xleft, xright, nxhet);
    }
    Arrays.sort(xgrid);
    if(yg != null) {
      ygrid = yg;
      nyhet = ygrid.length;
    } else {
      ygrid = computeGrid(yleft, yright, nyhet);
    }
    Arrays.sort(ygrid);
    if(zg != null) {
      zgrid = zg;
      nzhet = zgrid.length;
    } else {
      zgrid = computeGrid(zleft, zright, nzhet);
    }
    Arrays.sort(zgrid);

    // test the end points against the grid - the only time this matters is if
    // the grid lines are specified
    if(xleft.doubleValue() != xgrid[0].doubleValue()) xgrid[0] = xleft;
    if(xright.doubleValue() != xgrid[nxhet].doubleValue())
            xgrid[nxhet] = xright;
    if(yleft.doubleValue() != ygrid[0].doubleValue()) ygrid[0] = yleft;
    if(yright.doubleValue() != ygrid[nyhet].doubleValue())
            ygrid[nyhet] = yright;
    if(zleft.doubleValue() != zgrid[0].doubleValue()) zgrid[0] = zleft;
    if(zright.doubleValue() != zgrid[nzhet].doubleValue())
            zgrid[nzhet] = zright;

    // set up the computational parameters
    done = false;
    ok = false;

    // set up the value to return
    value = new Double[9];

    // set up the thread and run it to compute the average value for the data.
    computeThread = new Thread(this);
    computeThread.start();

  }

  public void run() {
    computeCellCenter();
    if(testData()) {
      computeCellCenter();
      if(computeValue()) {
        ok = true;
      } else {
        ok = false;
      }
    } else {
      ok = false;
    }
    done = true;
    computeThread = null;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for simple averaging                                         *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public boolean testData() {
    for (int k=1;k<=nzhet;k++) {
      int k2 = ( k - 1 ) * nxnyhet;
      for (int j=1;j<=nyhet;j++) {
        int j2 = ( j - 1 ) * nxhet + k2;
        for (int i=1;i<=nxhet;i++) {
          int index = i + j2 - 1;
          for(int l=0; l<9; l++) {
            if(data[index][l] == null) {
              System.out.println("AperiodicWaveletAverage: data test failure:"
                                + "\n   x index =       " + i
                                + "\n   y index =       " + j
                                + "\n   z index =       " + k
                                + "\n   spatial index = " + index
                                + "\n   component index = " + l);
              return false;
            }
          }
        }
      }
    }
    return true;
  }

  public boolean computeValue() {

    // test for an appropriate grid size - some power of two in each coordinate
    // direction
    if(!testNumberOfBlocks()) return false;

    // at this point we have a dyadic grid and can proceed with the calculation
    // loop over the levels
    int xlimVal = nxhet / 2; 
    if(nxhet == 1) xlimVal = 1;
    int ylimVal = nyhet / 2; 
    if(nyhet == 1) ylimVal = 1;
    int zlimVal = nzhet / 2; 
    if(nzhet == 1) zlimVal = 1;
    int skipBy2 = 2;
    for(int l=0; l<powOfTwo; l++) {
      for(int k=0; k<zlimVal; k++) {
        int k2 = skipBy2 * k * nxhet * nyhet; 
        for(int j=0; j<ylimVal; j++) {
          int j2 = skipBy2 * j * nxhet; 
          for(int i=0; i<xlimVal; i++) {
            int index = skipBy2 * i + j2;

            // pull off the eight tensor values for the 2x2 block currently
            // being considered
System.out.println("index:    " + index);
            double [][] tensI = setTensor(index);
            int indshft = index + skipBy2 / 2;
            double [][] tensII = setTensor(indshft);
            indshft = index + skipBy2 * nxhet / 2;
            double [][] tensIII = setTensor(indshft);
            indshft = index + skipBy2 / 2 + skipBy2 * nxhet / 2;
            double [][] tensIV = setTensor(indshft);
            indshft = index + skipBy2 * nxhet * nyhet / 2;
            double [][] tensV = setTensor(indshft);
System.out.println("indshft:  " + indshft);
            indshft = index + skipBy2 / 2 + skipBy2 * nxhet * nyhet / 2;
            double [][] tensVI = setTensor(indshft);
System.out.println("indshft:  " + indshft);
            indshft = index + skipBy2 * nxhet / 2 + skipBy2 * nxhet * nyhet / 2;
            double [][] tensVII = setTensor(indshft);
System.out.println("indshft:  " + indshft);
            indshft = index + skipBy2 / 2 + skipBy2 * nxhet / 2
                    + skipBy2 * nxhet * nyhet / 2;
            double [][] tensVIII = setTensor(indshft);
System.out.println("indshft:  " + indshft);

            // compute some parameters that we need to fill out the solution

            // average values
            double kxxbar = tensI[0][0] + tensII[0][0] + tensIII[0][0]
                          + tensIV[0][0] + tensV[0][0] + tensVI[0][0]
                          + tensVII[0][0] + tensVIII[0][0];
            double kxybar = tensI[0][1] + tensII[0][1] + tensIII[0][1]
                          + tensIV[0][1] + tensV[0][1] + tensVI[0][1]
                          + tensVII[0][1] + tensVIII[0][1];
            double kxzbar = tensI[0][2] + tensII[0][2] + tensIII[0][2]
                          + tensIV[0][2] + tensV[0][2] + tensVI[0][2]
                          + tensVII[0][2] + tensVIII[0][2];
            double kyybar = tensI[1][1] + tensII[1][1] + tensIII[1][1]
                          + tensIV[1][1] + tensV[1][1] + tensVI[1][1]
                          + tensVII[1][1] + tensVIII[1][1];
            double kyzbar = tensI[1][2] + tensII[1][2] + tensIII[1][2]
                          + tensIV[1][2] + tensV[1][2] + tensVI[1][2]
                          + tensVII[1][2] + tensVIII[1][2];
            double kzzbar = tensI[2][2] + tensII[2][2] + tensIII[2][2]
                          + tensIV[2][2] + tensV[2][2] + tensVI[2][2]
                          + tensVII[2][2] + tensVIII[2][2];

            // the p values
            double kxyp = tensI[0][1] - tensII[0][1] - tensIII[0][1]
                        + tensIV[0][1] + tensV[0][1] - tensVI[0][1]
                        - tensVII[0][1] + tensVIII[0][1];
            double kxzp = tensI[0][2] - tensII[0][2] + tensIII[0][2]
                        - tensIV[0][2] - tensV[0][2] + tensVI[0][2]
                        - tensVII[0][2] + tensVIII[0][2];
            double kyzp = tensI[1][2] + tensII[1][2] - tensIII[1][2]
                        - tensIV[1][2] - tensV[1][2] - tensVI[1][2]
                        + tensVII[1][2] + tensVIII[1][2];

            // the tilde values
            double kxxt = ( tensI[0][0] - tensII[0][0] + tensIII[0][0]
                        - tensIV[0][0] + tensV[0][0] - tensVI[0][0]
                        + tensVII[0][0] - tensVIII[0][0] ) / 16.0;
            double kxyt = ( tensI[0][1] + tensII[0][1] - tensIII[0][1]
                        - tensIV[0][1] + tensV[0][1] + tensVI[0][1]
                        - tensVII[0][1] - tensVIII[0][1] ) / 16.0;
            double kxzt = ( tensI[0][2] + tensII[0][2] + tensIII[0][2]
                        + tensIV[0][2] - tensV[0][2] - tensVI[0][2]
                        - tensVII[0][2] - tensVIII[0][2] ) / 16.0;
            double kyyt = ( tensI[1][1] + tensII[1][1] - tensIII[1][1]
                        - tensIV[1][1] + tensV[1][1] + tensVI[1][1]
                        - tensVII[1][1] - tensVIII[1][1] ) / 16.0;
            double kyzt = ( tensI[1][2] + tensII[1][2] + tensIII[1][2]
                        + tensIV[1][2] - tensV[1][2] - tensVI[1][2]
                        - tensVII[1][2] - tensVIII[1][2] ) / 16.0;
            double kzzt = ( tensI[2][2] + tensII[2][2] + tensIII[2][2]
                        + tensIV[2][2] - tensV[2][2] - tensVI[2][2]
                        - tensVII[2][2] - tensVIII[2][2] ) / 16.0;

            // the hat values
            double kxyh = kxyp / 16.0;
            double kxzh = ( tensI[0][2] - tensII[0][2] + tensIII[0][2]
                        - tensIV[0][2] + tensV[0][2] - tensVI[0][2]
                        + tensVII[0][2] - tensVIII[0][2] ) / 16.0;
            double kyzh = ( tensI[1][2] + tensII[1][2] - tensIII[1][2]
                        - tensIV[1][2] + tensV[1][2] + tensVI[1][2]
                        - tensVII[1][2] - tensVIII[1][2] ) / 16.0;

            // compute the coefficients of the solutions of the local problem in
            // the homogeniztion
            double denom = 2.0 * ( kxzp * kxzp * kyybar
                                 - 2.0 * kxyp * kxzp * kyzp
                                 + kxxbar * kyzp * kyzp
                                 + kzzbar * kxyp * kxyp
                                 - kxxbar * kyybar * kzzbar );

            double num1 = kxyp * kxyt * kxzp - kxyp * kxyp * kxzt
                        - kxxt * kxzp * kyybar + kxxbar * kxzt * kyybar
                        + kxxt * kxyp * kyzp - kxxbar * kxyt * kyzp;

            double num2 = - kxyt * kxzp * kxzp + kxyp * kxzp * kxzt
                        + kxxt * kxzp * kyzp - kxxbar * kxzt * kyzp
                        - kxxt * kxyp * kzzbar + kxxbar * kxyt * kzzbar;

            double num3 = - kxzp * kxzt * kyybar + kxyt * kxzbar * kyzbar
                        + kxybar * kxzt * kyzbar - kxxt * kyzp * kyzp
                        - kxyp * kxyt * kzzbar + kxxt * kyybar * kzzbar;

            double num4 = - kxyt * kxzp * kyybar + kxyp * kxzp * kyyt
                        + kxyp * kxyt * kyzp - kxxbar * kyyt * kyzbar
                        - kxyp * kxyp * kyzt + kxxbar * kyybar * kyzt;

            double num5 = - kxzp * kxzp * kyyt + kxyt * kxzp * kyzp
                        + kxyp * kxzp * kyzt - kxxbar * kyzp * kyzt
                        - kxyp * kxyt * kzzbar + kxxbar * kyyt * kzzbar;

            double num6 = kxzp * kyyt * kyzp - kxyt * kyzp * kyzp
                        - kxzp * kyybar * kyzt + kxyp * kyzp * kyzt
                        + kxyt * kyybar * kzzbar - kxyp * kyyt * kzzbar;

            double num7 = - kxzp * kxzt * kyybar + kxyp * kxzt * kyzp
                        + kxyp * kxzp * kyzt - kxxbar * kyzp * kyzt
                        - kxyp * kxyp * kzzt + kxxbar * kyybar * kzzt;

            double num8 = kxzp * kxzt * kyzp - kxyp * kxzp * kyzt
                        - kxyp * kxzt * kzzbar + kxxbar * kyzt * kzzbar
                        + kxyp * kxzp * kzzt - kxxbar * kyzp * kzzt;

            double num9 = - kxzt * kyzp * kyzp + kxzp * kyzp * kyzt
                        + kxzt * kyybar * kzzbar - kxyp * kyzt * kzzbar
                        - kxzp * kyybar * kzzt + kxyp * kyzp * kzzt;

            // compute the coefficients
            double alpha12 = num1 / denom;
            double alpha13 = num2 / denom;
            double alpha14 = num3 / denom;
            double alpha22 = num4 / denom;
            double alpha23 = num5 / denom;
            double alpha24 = num6 / denom;
            double alpha32 = num7 / denom;
            double alpha33 = num8 / denom;
            double alpha34 = num9 / denom;

            // set up the tensor that results from the local 2x2 problem
            double [][] kout = new double[3][3];
            kout[0][0] = 0.125 * kxxbar
                + 4.0 * ( alpha12 * kxzt + alpha13 * kxyt + alpha14 * kxxt );
            kout[0][1] = 0.125 * kxybar
                + 4.0 * ( alpha22 * kxzt + alpha23 * kxyt + alpha24 * kxxt );
            kout[0][2] = 0.125 * kxzbar
                + 4.0 * ( alpha32 * kxzt + alpha33 * kxyt + alpha34 * kxxt );
            kout[1][0] = kout[0][1];
            kout[1][1] = 0.125 * kyybar
                + 4.0 * ( alpha22 * kyzt + alpha23 * kyyt + alpha24 * kxyh );
            kout[1][2] = 0.125 * kyzbar
                + 4.0 * ( alpha32 * kyzt + alpha33 * kyyt + alpha34 * kxyh );
            kout[2][0] = kout[0][2];
            kout[2][1] = kout[1][2];
            kout[2][2] = 0.125 * kzzbar
                + 4.0 * ( alpha32 * kzzt + alpha33 * kyzh + alpha34 * kxzh );

            // reset the appropriate cell to hold the average value at this
            // level
            getTensor(index, kout);

          }
        }
      }
      xlimVal = xlimVal / 2;
      ylimVal = ylimVal / 2;
      skipBy2 = 2 * skipBy2;
    }

    for(int l=0; l<9; l++) value[l] = data[0][l];

    notifyObservers();
    return true;
  }

  private double [][] setTensor(int index) {
    if(index < 0 || index >= nnhet) return null;
    double [][] tens = new double[3][3];
    tens[0][0] = data[index][0].doubleValue();
    tens[0][1] = data[index][1].doubleValue();
    tens[0][2] = data[index][2].doubleValue();
    tens[1][0] = data[index][3].doubleValue();
    tens[1][1] = data[index][4].doubleValue();
    tens[1][2] = data[index][5].doubleValue();
    tens[2][0] = data[index][6].doubleValue();
    tens[2][1] = data[index][7].doubleValue();
    tens[2][2] = data[index][8].doubleValue();
    return tens;
  }

  private void getTensor(int index, double [][] tens) {
    if(index < 0 || index >= nnhet) return;
    data[index][0] = new Double(tens[0][0]);
    data[index][1] = new Double(tens[0][1]);
    data[index][2] = new Double(tens[0][2]);
    data[index][3] = new Double(tens[1][0]);
    data[index][4] = new Double(tens[1][1]);
    data[index][5] = new Double(tens[1][2]);
    data[index][6] = new Double(tens[2][0]);
    data[index][7] = new Double(tens[2][1]);
    data[index][8] = new Double(tens[2][2]);
  }

  public boolean testNumberOfBlocks() {
    if(nzhet > 1) return false;
    if(nxhet != nyhet) return false;
    int ival = nxhet;
    int pow2 = 1;
    powOfTwo = 0;
    while(ival>1) {
      ival = ival / 2;
      pow2 = 2 * pow2;
      powOfTwo++;
    }
    return true;
  }

  public Double [] computeGrid(Double xl, Double xr, int nx) {
    Double [] grid = null;
    grid = new Double [nx+1];
    double dx = (xr.doubleValue() - xl.doubleValue()) / ((double) nx);
    grid[0] = xl;
    for(int i=1; i<nx; i++) {
      grid[i] = new Double(grid[i-1].doubleValue() + dx);
    }
    grid[nx] = xr;
    return grid;
  }

  public void computeCellCenter() {
    xcenter = new Double(0.5 * ( xleft.doubleValue() + xright.doubleValue() ));
    ycenter = new Double(0.5 * ( yleft.doubleValue() + yright.doubleValue() ));
    zcenter = new Double(0.5 * ( zleft.doubleValue() + zright.doubleValue() ));
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for getting the computed value at an intermediate point      *
   *                                                                         *
   *     note that a capital "F" indicates a return type of Double while a   *
   *          small "f' indicates a return type of double                    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getFunctionValue(Double x, Double y, Double z) {
    return getFunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public Double [] getFunctionValue(double x, double y, double z) {
    return value;
  }

  public double [] getfunctionValue(Double x, Double y, Double z) {
    return getfunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public double [] getfunctionValue(double x, double y, double z) {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the computed value                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getDoubleValue() {
    return value;
  }

  public double [] getdoubleValue() {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the internal variables - the most important is    *
   * probably getTag which returns the integer identification of the object  *
   * which must be set in the instantiation                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
 
  public Double [][] getData() {
    return data;
  }
 
  public int getNumberOfXBlocks() {
    return nxhet;
  }
 
  public int getNumberOfYBlocks() {
    return nyhet;
  }
 
  public int getNumberOfZBlocks() {
    return nzhet;
  }
 
  public int getTotalBlocks() {
    return nnhet;
  }

  public Double getXCenter() {
    return xcenter;
  }

  public Double getYCenter() {
    return ycenter;
  }

  public Double getZCenter() {
    return zcenter;
  }

  public double getxCenter() {
    return xcenter.doubleValue();
  }

  public double getyCenter() {
    return ycenter.doubleValue();
  }

  public double getzCenter() {
    return zcenter.doubleValue();
  }

  public Double getXLeft() {
    return xleft;
  }

  public double getxLeft() {
    return xleft.doubleValue();
  }

  public Double getXRight() {
    return xright;
  }

  public double getxRight() {
    return xright.doubleValue();
  }

  public Double getYLeft() {
    return yleft;
  }

  public double getyLeft() {
    return yleft.doubleValue();
  }

  public Double getYRight() {
    return yright;
  }

  public double getyRight() {
    return yright.doubleValue();
  }

  public Double getZLeft() {
    return zleft;
  }

  public double getzLeft() {
    return zleft.doubleValue();
  }

  public Double getZRight() {
    return zright;
  }

  public double getzRight() {
    return zright.doubleValue();
  }
 
  public int getTag() {
    return tag;
  }
 
  public boolean isValueOK() {
    return ok;
  }
 
  public boolean getStatus() {
    return done;
  }

  public static String digitFilter(Double value, int n) {
    double dval = value.doubleValue();
    return digitFilter(dval, n);
  }

  public static String digitFilter(double value, int n) {
    String integerPart = "0";
    String dot = ".";
    String fractionPart = "0";
    String exponent = "";
    String t = null;
    String s = Double.toString(value);
    StringTokenizer st = new StringTokenizer(s, " .eE");
    if(st.hasMoreTokens()) integerPart = st.nextToken();
    if(st.hasMoreTokens()) t = st.nextToken();
    if(t != null) {
      fractionPart = t;
      if(t.length() > n) fractionPart = t.substring(0,n);
    }
    if(st.hasMoreTokens()) exponent = st.nextToken();
    if(exponent.length() > 0) exponent = "E" + exponent;
    return integerPart + dot + fractionPart + exponent;
  }
 
  public String toString() {
    String tabChar = "\t";
    String s = "";
    s = "\nString output for the AperiodicWavelet Average Object\n\n";
    s = s + "  |  ";
    if(value[0] != null) {
      s = s + tabChar + "axx= " + digitFilter(value[0], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[1] != null) {
      s = s + tabChar + "axy= " + digitFilter(value[1], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[2] != null) {
      s = s + tabChar + "axz= " + digitFilter(value[2], 5);
    } else {
      s = s + tabChar + "null";
    }
    s = s + "  |\n";
    s = s + "  |  ";
    if(value[3] != null) {
      s = s + tabChar + "ayx= " + digitFilter(value[3], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[4] != null) {
      s = s + tabChar + "ayy= " + digitFilter(value[4], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[5] != null) {
      s = s + tabChar + "ayz= " + digitFilter(value[5], 5);
    } else {
      s = s + tabChar + "null";
    }
    s = s + "  |\n";
    s = s + "  |  ";
    if(value[6] != null) {
      s = s + tabChar + "azx= " + digitFilter(value[6], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[7] != null) {
      s = s + tabChar + "azy= " + digitFilter(value[7], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[8] != null) {
      s = s + tabChar + "azz= " + digitFilter(value[8], 5);
    } else {
      s = s + tabChar + "null";
    }
    s = s + "  |\n";
    return s;
  }
 
  private String tensorString(double [][] tens) {
    String tabChar = "\t";
    String s = "";
    s = "\nString version of a tensor in an Average Object\n\n";
    s = s + "  |  ";
      s = s + tabChar + "axx= " + digitFilter(tens[0][0], 5);
      s = s + tabChar + "axy= " + digitFilter(tens[0][1], 5);
      s = s + tabChar + "axz= " + digitFilter(tens[0][2], 5);
    s = s + "  |\n";
    s = s + "  |  ";
      s = s + tabChar + "ayx= " + digitFilter(tens[1][0], 5);
      s = s + tabChar + "ayy= " + digitFilter(tens[1][1], 5);
      s = s + tabChar + "ayz= " + digitFilter(tens[1][2], 5);
    s = s + "  |\n";
    s = s + "  |  ";
      s = s + tabChar + "azx= " + digitFilter(tens[2][0], 5);
      s = s + tabChar + "azy= " + digitFilter(tens[2][1], 5);
      s = s + tabChar + "azz= " + digitFilter(tens[2][2], 5);
    s = s + "  |\n";
    return s;
  }
 
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    main method for testing the interface                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public static void main(String args[]) {
    int i, j, k, l;
    int nx = 4;
    int ny = 4;
    int nz = 1;
    int ntot = nx * ny * nz;
    // a first simple test on a single block
    AperiodicWaveletAverage awa;
    Double [][] data = new Double[ntot][9];
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(0.0);
    for(i=0; i<ntot; i++) {
      data[i][0] = new Double(i+1.0);
      data[i][4] = new Double(i+1.0);
      data[i][8] = new Double(i+1.0);
    }
    try {
      awa = new AperiodicWaveletAverage(data, nx, ny, nz, 1);
      System.out.println("getData() test:   ");
      data = awa.getData();
// uncomment the next 5 lines to see the input data as the code sees it
//      for(l=0; l<9; l++) {
//        for(i=0; i<ntot; i++) {
//           System.out.println("Data Value: " + i + " " + data[i][l]);
//        }
//      }
      while(!awa.getStatus()) {
        System.out.println("isValueOK() test:   " + awa.isValueOK());
        System.out.println("getStatus() test:   " + awa.getStatus());
        System.out.println("   ");
        // sleep the process and test again
        try {Thread.sleep(100);
        } catch(InterruptedException ie) {
          break;
        }
      }
      System.out.println("Double Output: " + awa.toString());
      //System.out.println("double Output: " + awa.getdoubleValue());
      System.out.println("isValueOK() test:   " + awa.isValueOK());
      System.out.println("getStatus() test:   " + awa.getStatus());
      nx = awa.getNumberOfXBlocks();
      System.out.println("getNumberOfXBlocks() test:   " + nx);
      ny = awa.getNumberOfYBlocks();
      System.out.println("getNumberOfYBlocks() test:   " + ny);
      nz = awa.getNumberOfZBlocks();
      System.out.println("getNumberOfZBlocks() test:   " + nz);
      System.out.println("getXLeft() test:             " + awa.getXLeft());
      System.out.println("getxLeft() test:             " + awa.getxLeft());
      System.out.println("getXRight() test:            " + awa.getXRight());
      System.out.println("getxRight() test:            " + awa.getxRight());
      System.out.println("getYLeft() test:             " + awa.getYLeft());
      System.out.println("getyLeft() test:             " + awa.getyLeft());
      System.out.println("getYRight() test:            " + awa.getYRight());
      System.out.println("getyRight() test:            " + awa.getyRight());
      System.out.println("getZLeft() test:             " + awa.getZLeft());
      System.out.println("getzLeft() test:             " + awa.getzLeft());
      System.out.println("getZRight() test:            " + awa.getZRight());
      System.out.println("getzRight() test:            " + awa.getzRight());
      System.out.println("getXCenter() test:           " + awa.getXCenter());
      System.out.println("getYCenter() test:           " + awa.getYCenter());
      System.out.println("getZCenter() test:           " + awa.getZCenter());
      System.out.println("getxCenter() test:           " + awa.getxCenter());
      System.out.println("getyCenter() test:           " + awa.getyCenter());
      System.out.println("getzCenter() test:           " + awa.getzCenter());
      ntot = awa.getTotalBlocks();
      System.out.println("getTotalBlocks() test:   " + ntot);
      System.out.println("getTag() test:   " + awa.getTag());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }
  }

  // local variables
  private Double [] value;
  private Double [][] data;
  private int nxhet;
  private int nyhet;
  private int nzhet;
  private int nxnyhet;
  private int nnhet;
  private int powOfTwo = 0;
  private Double xleft;
  private Double xright;
  private Double yleft;
  private Double yright;
  private Double zleft;
  private Double zright;
  private Double [] xgrid;
  private Double [] ygrid;
  private Double [] zgrid;
  private Double xcenter;
  private Double ycenter;
  private Double zcenter;

  // a thread variable for running the neccesary averaging
  private Thread computeThread;
  private boolean done = false;
  private boolean ok = false;

  // the following integer is a tag to identify the object for use by the
  // object which instantitates one of these averaging things
  private int tag;

}

