/*
 * SimulatorGUI class: BVidalFracCodeDispAverage
 *
 */

// the following list identifies the steps you need to take care of to build
// an averaging object for your favorite method. the steps can be searched for
// using your editor. there is a comment with the step name (e;g;, Step 1) at
// each appropriate location in the code.
//
//   Step 1. Identify the location of the executable image and the names of
//           the input and output files that the executable expects when it
//           is run. these are listed close to the top of this file.
//   Step 2. (optional) Set up methods for setting and getting the location
//           of the executable and input and output file names. If you want
//           to hardwire these in one place you do not need this step.
//   Step 3. Set the name of the working directories to be generated by the
//           java code in which to run your executable. the input data file
//           and the shell script automatically generated by this code will be
//           thrown into directories with this name and the tag value at the
//           end. so each process (the code is multithreaded) will have a
//           unique directory in which to work.
//   Step 4. Set up the formatting for the input file to the executable. this
//           is dependent on the format needed by the executable. in the case
//           below a separate method was developed to produce an output file
//           in the correct format.
//   Step 5. Set up a means for getting the averaged values back into this
//           code. The input needs to read in a total of 9 values for this
//           code. So, you need to make sure that the 9 values are set in the
//           input. In some cases, you may not have the information in the
//           code. For example, If your code is a 2D code, you may only have
//           the two component direction values. in this case you might want
//           to set the third component values to 0 on the off diagonal and to
//           some appropriate value on the diagonal, say 1.
//

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import java.util.Observable;
import java.util.Random;
import java.util.StringTokenizer;

/*
 * this is an object that will compute the arithmetic average and store the
 * appropriate value for retrieval
 */
public class BVidalFracCodeDispAverage extends AverageObject {

  // Step 1.
  // Code location and input and output file names set the input and output
  // file names - these are default names that may be overridden using the set
  // methods below.
  private static String userHome = System.getProperty("user.home");
  // the first two lines assume that the package has been installed in the
  // users home directory. if this is not the case this must be changed
  // accordingly
  private static String homCodeLocation = userHome
                          + File.separator
                          + "/java/JHomogenizer/src/f90/BVidalFracCode/disp";
  // input and output file names from the source code
  private static String homCodeInputFileName = "dispInput.dat";
  private static String homCodeOutputFileName = "dispOutput.dat";

  // uniform grid on the unit cubic constructor for the class
  public BVidalFracCodeDispAverage(Double [][] d, int nx, int ny, int nz, int t)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         null, null, null);
  }

  // uniform grid on a specified region constructor for the class
  public BVidalFracCodeDispAverage(Double [][] d, int nx, int ny, int nz, int t,
                           Double xl, Double xr,
                           Double yl, Double yr,
                           Double zl, Double zr)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         xl, xr, yl, yr, zl, zr,
         null, null, null);
  }

  // specified grid constructor for the class
  public BVidalFracCodeDispAverage(Double [][] d, int nx, int ny, int nz, int t,
                           Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         xg, yg, zg);
  }

  // full constructor for the class
  public BVidalFracCodeDispAverage(Double [][] d, int nx, int ny, int nz, int t,
                           Double xl, Double xr,
                           Double yl, Double yr,
                           Double zl, Double zr,
                           Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    data = d;
    nxhet = nx;
    nyhet = ny;
    nzhet = nz;
    tag = t;
    nxnyhet = nxhet * nyhet;
    nnhet = nxnyhet * nzhet;
    if(data.length != nnhet) throw
         new BadArraySizeException("Array the wrong size in BVidalFracCodeDispAverage");

    // set up the grid
    if(xl != null) {
      xleft = xl;
    } else {
      xleft = new Double(0.0);
    }
    if(xr != null) {
      xright= xr;
    } else {
      xright = new Double(1.0);
    }
    if(yl != null) {
      yleft = yl;
    } else {
      yleft = new Double(0.0);
    }
    if(yr != null) {
      yright= yr;
    } else {
      yright = new Double(1.0);
    }
    if(zl != null) {
      zleft = zl;
    } else {
      zleft = new Double(0.0);
    }
    if(zr != null) {
      zright= zr;
    } else {
      zright = new Double(1.0);
    }

    // make sure the end points are ok
    if(xleft.doubleValue() > xright.doubleValue()) {
      Double temp = xleft;
      xleft = xright;
      xright = temp;
    }
    if(yleft.doubleValue() > yright.doubleValue()) {
      Double temp = yleft;
      yleft = yright;
      yright = temp;
    }
    if(zleft.doubleValue() > zright.doubleValue()) {
      Double temp = zleft;
      zleft = zright;
      zright = temp;
    }

    // set or compute the grid lines in the domain
    if(xg != null) {
      xgrid = xg;
      nxhet = xgrid.length;
    } else {
      xgrid = computeGrid(xleft, xright, nxhet);
    }
    Arrays.sort(xgrid);
    if(yg != null) {
      ygrid = yg;
      nyhet = ygrid.length;
    } else {
      ygrid = computeGrid(yleft, yright, nyhet);
    }
    Arrays.sort(ygrid);
    if(zg != null) {
      zgrid = zg;
      nzhet = zgrid.length;
    } else {
      zgrid = computeGrid(zleft, zright, nzhet);
    }
    Arrays.sort(zgrid);

    // test the end points against the grid - the only time this matters is if
    // the grid lines are specified
    if(xleft.doubleValue() != xgrid[0].doubleValue()) xgrid[0] = xleft;
    if(xright.doubleValue() != xgrid[nxhet].doubleValue())
            xgrid[nxhet] = xright;
    if(yleft.doubleValue() != ygrid[0].doubleValue()) ygrid[0] = yleft;
    if(yright.doubleValue() != ygrid[nyhet].doubleValue())
            ygrid[nyhet] = yright;
    if(zleft.doubleValue() != zgrid[0].doubleValue()) zgrid[0] = zleft;
    if(zright.doubleValue() != zgrid[nzhet].doubleValue())
            zgrid[nzhet] = zright;

    // set up the computational parameters
    done = false;
    ok = false;

    // set up the value to return
    value = new Double[9];

    // set up the thread and run it to compute the average value for the data.
    computeThread = new Thread(this);
    computeThread.start();

  }

  public void run() {
    computeCellCenter();
    if(testData()) {
      computeCellCenter();
      if(computeValue()) {
        ok = true;
      } else {
        ok = false;
      }
    } else {
      ok = false;
    }
    done = true;
    computeThread = null;
  }

  // Step 2.
  // Methods for setting and getting the executable location and the names of
  // the input and  output files. you might want to use these from another
  // piece of code.
  public void setBVidalFracCodeDispLocation(String s) {
    homCodeLocation = s;
  }
  public void setBVidalFracCodeDispInputFileName(String s) {
    homCodeInputFileName = s;
  }
  public void setBVidalFracCodeDispOutputFileName(String s) {
    homCodeOutputFileName = s;
  }
  public String getBVidalFracCodeDispLocation() {
    return homCodeLocation;
  }
  public String getBVidalFracCodeDispInputFileName() {
    return homCodeInputFileName;
  }
  public String getBVidalFracCodeDispOutputFileName() {
    return homCodeOutputFileName;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for simple averaging                                         *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public boolean testData() {
    for (int k=1;k<=nzhet;k++) {
      int k2 = ( k - 1 ) * nxnyhet;
      for (int j=1;j<=nyhet;j++) {
        int j2 = ( j - 1 ) * nxhet + k2;
        for (int i=1;i<=nxhet;i++) {
          int index = i + j2 - 1;
          for(int l=0; l<9; l++) {
            if(data[index][l] == null) {
              System.out.println("BVidalFracCodeDispAverage: data test failure:"
                                + "\n   x index =       " + i
                                + "\n   y index =       " + j
                                + "\n   z index =       " + k
                                + "\n   spatial index = " + index
                                + "\n   component index = " + l);
              return false;
            }
          }
        }
      }
    }
    return true;
  }

  public boolean computeValue() {
    String fileSeparator = File.separator;
    int j2, k2, index;

    // check to make sure that the software location and input and output
    // file information has been set
    if(homCodeLocation == null) return false;
    if(homCodeInputFileName == null) return false;
    if(homCodeOutputFileName == null) return false;

    // Step 3.
    // this line sets the name of the shell script that will be used to
    // execute the native code
    String shellName = "homshell";
    // first, we need to set up a file to place the data in - note that this
    // sets up a working directory in which the executable does its own thing.
    // that way each object has a disjoint work space for its own local
    // problem. keep in mind that the code is multithreaded and that if there
    // are multiple cells being homogenized using your executable, each one
    // will be running on its own and needs its own space.
    String fileString = "tmp" + File.separator + "bvdispData"
                      + new Integer(getTag()).toString().trim();
    File f = new File(fileString);
    if(!f.exists()) f.mkdirs();

    // Step 4.
    // the following code is in place for testing the original homcode from
    // a past life - we need a few data files. these files are created in the
    // appropriate place using the following method. if you like, you can
    // embed the code right here and not invoke a method.
    bvDataOutput(fileString);

    // now at this point we have the data out there - so we need to send the
    // data through the code
    File fd = new File(fileString, shellName);
    try{
      FileWriter fw = new FileWriter(fd);
      String line;
      try{

        line = "# shell script to run BVidalFracCodeDisp from JHomogenizer\n";
        fw.write(line, 0, line.length());
        line = "# generated automatically - can be erased at any time\n\n";
        fw.write(line, 0, line.length());

        File fp = fd.getParentFile();
        line = "# change to the directory in which to run the code\n"
             + "cd "
             + fp.getAbsolutePath()
             + "\n\n";
        fw.write(line, 0, line.length());

        line = "# run the code\n"
             + homCodeLocation
             + "\n\n";
        fw.write(line, 0, line.length());

      } catch(IOException ioe) {
        return false;
      }
      fw.flush();
      fw.close();

    } catch(IOException ioe) {
      return false;
    }

    // we need to tell the system that the shell script is exectuable
    ExternalProcess plusXProcess = new ExternalProcess(null, 2000);
    String command = "chmod +x "
                   + fileString
                   + File.separator
                   + shellName;
    plusXProcess.setCommand(command);
    plusXProcess.runCommand();
    // test to see if the code is done running or not
    while(plusXProcess.getStatus()) {
      try {
        Thread.sleep(100);
      } catch(InterruptedException ie) {}
    }

    // execute the command line
    ExternalProcess homCodeProcess = new ExternalProcess(null, 2000);
    command = "/bin/csh ."
            + File.separator
            + fileString
            + File.separator
            + shellName;
    homCodeProcess.setCommand(command);
    homCodeProcess.runCommand();

    // test to see if the code is done running or not
    while(homCodeProcess.getStatus()) {
      try {
        Thread.sleep(1000);
      } catch(InterruptedException ie) {}
    }

    // now we are ready to pull the data back into the deal if the code
    // ran correctly
    if(homCodeProcess.getExitValue() != 0) return false;

    // the following code is in place for testing the original homcode from
    // a past life - we need a few data files
    //
    if(!bvDataInput(fileString)) return false;
    
    notifyObservers();
    return true;

  }
 
  // Step 4.
  // the following method creates a total of 4 input files for the original
  // HomCode

//10
//#taille_cellule_selon_y
//10
//#taille_matrice_selon_x
//8
//#taille_cellule_selon_y
//8
//#raffinement_selon_x
//2
//#raffinement_selon_y
//2
//#permeabilite_dans_fracture
//1e-7
//#permeabilite_macro
//2e-8
//#diffusion_effective_dans_fracture
//1e-11
//#diffusion_effective_dans_matrice
//1e-12
//#gradient_de_charge_composante_horizontale
//5e-4
//#gradient_de_charge_composante_verticale
//5e-4

  // end of the input file
  // input routine
  public boolean bvDataOutput(String fileString) {
    int index;

    // THE INPUT FILE
    File fd = new File(fileString, homCodeInputFileName);
    try{
      FileWriter fw = new FileWriter(fd);
      String line;
      try{
        line = "#DONNEES_POUR_CAS_FRACTURE_AVEC_DIFFUSION_ET_CONVECTION_EN_2D";
        fw.write(line, 0, line.length());

        line = "#taille_cellule_selon_x";
        fw.write(line, 0, line.length());
        line = new Integer(nxhet).toString();
        fw.write(line, 0, line.length());

        line = "#taille_cellule_selon_y";
        fw.write(line, 0, line.length());
        line = new Integer(nyhet).toString();
        fw.write(line, 0, line.length());

        line = "#taille_matrice_selon_x"; 
        fw.write(line, 0, line.length());
        line = new Integer(nxhet-2).toString();
        fw.write(line, 0, line.length());

        line = "#taille_matrice_selon_y"; 
        fw.write(line, 0, line.length());
        line = new Integer(nyhet-2).toString();
        fw.write(line, 0, line.length());

        line = "#raffinement_selon_x";
        fw.write(line, 0, line.length());
        line = new Integer(1).toString();
        fw.write(line, 0, line.length());

        line = "#raffinement_selon_y";
        fw.write(line, 0, line.length());
        line = new Integer(1).toString();
        fw.write(line, 0, line.length());

        line = "#permeabilite_dans_fracture";
        fw.write(line, 0, line.length());
double pinf = 1e-7;
        line = new Double(pinf).toString();
        fw.write(line, 0, line.length());

        line = "#permeabilite_macro";
        fw.write(line, 0, line.length());
double pm = 2e-8;
        line = new Double(pinf).toString();
        fw.write(line, 0, line.length());

        line = "#diffusion_effective_dans_fracture";
        fw.write(line, 0, line.length());
double pef = 1e-11;
        line = new Double(pef).toString();
        fw.write(line, 0, line.length());

        line = "#diffusion_effective_dans_matrice";
        fw.write(line, 0, line.length());
double diffh = 1e-12;
        line = new Double(diffh).toString();
        fw.write(line, 0, line.length());

        line = "#gradient_de_charge_composante_horizontale";
        fw.write(line, 0, line.length());
double gradh = 5e-4;
        line = new Double(gradh).toString();
        fw.write(line, 0, line.length());

        line = "#gradient_de_charge_composante_verticale";
        fw.write(line, 0, line.length());
double gradv = 5e-4;
        line = new Double(gradv).toString();
        fw.write(line, 0, line.length());

      } catch(IOException ioe) {
        return false;
      }
      fw.flush();
      fw.close();

    } catch(IOException ioe) {
      return false;
    }
 
    return true;

  }
 
  // Step 5.
  // this method pulls in the output from the executable
  public boolean bvDataInput(String fileString) {

    // read in the homogenized values
    File fd = new File(fileString, homCodeOutputFileName);
    try{
      FileReader fr = new FileReader(fd);
      String line;
      try{
        LineNumberReader lnr = new LineNumberReader(fr);
        line = lnr.readLine();
        line = lnr.readLine();
        line = lnr.readLine();
        // the first pieces read lines that qre not important
        // read the values of the tensor
        for(int l=0; l<9; l++) {
          line = lnr.readLine().trim();
          try {
            double test = Double.valueOf(line).doubleValue();
            value[l] = Double.valueOf(line);
          } catch(NumberFormatException nfe) {
            return false;
          }
        }
        lnr.close();
      } catch(IOException ioe) {
        return false;
      }
    } catch(FileNotFoundException fnfe) {
      return false;
    }

    return true;

  }

  public int mod(int i1, int i2) {
    return i1 - i2 * ((int) i1 / i2 );
  }

  public Double [] computeGrid(Double xl, Double xr, int nx) {
    Double [] grid = null;
    grid = new Double [nx+1];
    double dx = (xr.doubleValue() - xl.doubleValue()) / ((double) nx);
    grid[0] = xl;
    for(int i=1; i<nx; i++) {
      grid[i] = new Double(grid[i-1].doubleValue() + dx);
    }
    grid[nx] = xr;
    return grid;
  }

  public void computeCellCenter() {
    xcenter = new Double(0.5 * ( xleft.doubleValue() + xright.doubleValue() ));
    ycenter = new Double(0.5 * ( yleft.doubleValue() + yright.doubleValue() ));
    zcenter = new Double(0.5 * ( zleft.doubleValue() + zright.doubleValue() ));
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for getting the computed value at an intermediate point      *
   *                                                                         *
   *     note that a capital "F" indicates a return type of Double while a   *
   *          small "f' indicates a return type of double                    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getFunctionValue(Double x, Double y, Double z) {
    return getFunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public Double [] getFunctionValue(double x, double y, double z) {
    return value;
  }

  public double [] getfunctionValue(Double x, Double y, Double z) {
    return getfunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public double [] getfunctionValue(double x, double y, double z) {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the computed value                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getDoubleValue() {
    return value;
  }

  public double [] getdoubleValue() {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) {
      if(value[l] != null) {
        dval[l] = value[l].doubleValue();
      } else {
        dval[l] = -1.0;
      }
    }
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the internal variables - the most important is    *
   * probably getTag which returns the integer identification of the object  *
   * which must be set in the instantiation                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
 
  public Double [][] getData() {
    return data;
  }
 
  public int getNumberOfXBlocks() {
    return nxhet;
  }
 
  public int getNumberOfYBlocks() {
    return nyhet;
  }
 
  public int getNumberOfZBlocks() {
    return nzhet;
  }
 
  public int getTotalBlocks() {
    return nnhet;
  }

  public Double getXCenter() {
    return xcenter;
  }

  public Double getYCenter() {
    return ycenter;
  }

  public Double getZCenter() {
    return zcenter;
  }

  public double getxCenter() {
    return xcenter.doubleValue();
  }

  public double getyCenter() {
    return ycenter.doubleValue();
  }

  public double getzCenter() {
    return zcenter.doubleValue();
  }

  public Double getXLeft() {
    return xleft;
  }

  public double getxLeft() {
    return xleft.doubleValue();
  }

  public Double getXRight() {
    return xright;
  }

  public double getxRight() {
    return xright.doubleValue();
  }

  public Double getYLeft() {
    return yleft;
  }

  public double getyLeft() {
    return yleft.doubleValue();
  }

  public Double getYRight() {
    return yright;
  }

  public double getyRight() {
    return yright.doubleValue();
  }

  public Double getZLeft() {
    return zleft;
  }

  public double getzLeft() {
    return zleft.doubleValue();
  }

  public Double getZRight() {
    return zright;
  }

  public double getzRight() {
    return zright.doubleValue();
  }
 
  public int getTag() {
    return tag;
  }
 
  public boolean isValueOK() {
    return ok;
  }
 
  public boolean getStatus() {
    return done;
  }
 
  public String toString() {
    String s = "";
    s = "\nString output for the BVidalFracCodeDisp Average Object\n\n";
    s = s + " |  ";
    if(value[0] != null) {
      s = s + "axx= " + value[0].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[1] != null) {
      s = s + "axy= " + value[1].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[2] != null) {
      s = s + "axz= " + value[2].toString();
    } else {
      s = s + "null";
    }
    s = s + "  |\n";
    s = s + " |  ";
    if(value[3] != null) {
      s = s + "ayx= " + value[3].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[4] != null) {
      s = s + "ayy= " + value[4].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[5] != null) {
      s = s + "ayz= " + value[5].toString();
    } else {
      s = s + "null";
    }
    s = s + "  |\n";
    s = s + " |  ";
    if(value[6] != null) {
      s = s + "azx= " + value[6].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[7] != null) {
      s = s + "azy= " + value[7].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[8] != null) {
      s = s + "azz= " + value[8].toString();
    } else {
      s = s + "null";
    }
    s = s + "  |\n";
    return s;
  }

  public static String digitFilter(double value, int n) {
    String integerPart = "0";
    String dot = ".";
    String fractionPart = "0";
    String exponent = "";
    String t = null;

    String s = Double.toString(value);
    StringTokenizer st = new StringTokenizer(s, " .eE");
    if(st.hasMoreTokens()) integerPart = st.nextToken();
    if(st.hasMoreTokens()) t = st.nextToken();
    if(t != null) {
      fractionPart = t;
      if(t.length() > n) fractionPart = t.substring(0,n);
    }
    if(st.hasMoreTokens()) exponent = st.nextToken();
    if(exponent.length() > 0) exponent = "E" + exponent;

    return integerPart + dot + fractionPart + exponent;
  }
 
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    main method for testing the interface                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public static void main(String args[]) {
    String line;
    int i, j, k, l;
    int nx = 10;
    int ny = 10;
    int nz = 1;
    int ntot = nx * ny * nz;
    // a first simple test on a single block
    BVidalFracCodeDispAverage aa;
    Random r = new Random();
    Double [][] data = new Double[100][9];
    for(i=0; i<100; i++) for(l=0; l<9; l++) {
      data[i][0] = new Double(r.nextDouble());
      data[i][1] = new Double(0.0);
      data[i][2] = new Double(0.0);
      data[i][3] = new Double(0.0);
      data[i][4] = data[i][0];
      data[i][5] = new Double(0.0);
      data[i][6] = new Double(0.0);
      data[i][7] = new Double(0.0);
      data[i][8] = data[i][0];
    }
    try {
      aa = new BVidalFracCodeDispAverage(data, nx, ny, nz, 1);
      System.out.println("getData() test:   ");
      data = aa.getData();
      for(i=0; i<ntot; i++) {
        line = "Data Value: " + i;
        for(l=0; l<9; l++) line = line + " " + data[i][l];
        System.out.println(line);
      }
      while(!aa.getStatus()) {
        System.out.println("isValueOK() test:   " + aa.isValueOK());
        System.out.println("getStatus() test:   " + aa.getStatus());
        System.out.println("   ");
        // sleep the process and test again
        try {Thread.sleep(1000);
        } catch(InterruptedException ie) {
          break;
        }
      }
      System.out.println("Double Output: " + aa.toString());
      System.out.println("double Output: " + aa.getdoubleValue());
      System.out.println("isValueOK() test:   " + aa.isValueOK());
      System.out.println("getStatus() test:   " + aa.getStatus());
      nx = aa.getNumberOfXBlocks();
      System.out.println("getNumberOfXBlocks() test:   " + nx);
      ny = aa.getNumberOfYBlocks();
      System.out.println("getNumberOfYBlocks() test:   " + ny);
      nz = aa.getNumberOfZBlocks();
      System.out.println("getNumberOfZBlocks() test:   " + nz);
      System.out.println("getXLeft() test:             " + aa.getXLeft());
      System.out.println("getxLeft() test:             " + aa.getxLeft());
      System.out.println("getXRight() test:            " + aa.getXRight());
      System.out.println("getxRight() test:            " + aa.getxRight());
      System.out.println("getYLeft() test:             " + aa.getYLeft());
      System.out.println("getyLeft() test:             " + aa.getyLeft());
      System.out.println("getYRight() test:            " + aa.getYRight());
      System.out.println("getyRight() test:            " + aa.getyRight());
      System.out.println("getZLeft() test:             " + aa.getZLeft());
      System.out.println("getzLeft() test:             " + aa.getzLeft());
      System.out.println("getZRight() test:            " + aa.getZRight());
      System.out.println("getzRight() test:            " + aa.getzRight());
      System.out.println("getXCenter() test:           " + aa.getXCenter());
      System.out.println("getYCenter() test:           " + aa.getYCenter());
      System.out.println("getZCenter() test:           " + aa.getZCenter());
      System.out.println("getxCenter() test:           " + aa.getxCenter());
      System.out.println("getyCenter() test:           " + aa.getyCenter());
      System.out.println("getzCenter() test:           " + aa.getzCenter());
      ntot = aa.getTotalBlocks();
      System.out.println("getTotalBlocks() test:   " + ntot);
      System.out.println("getTag() test:   " + aa.getTag());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }
  }

  // local variables
  private Double [] value;
  private Double [][] data;
  private int nxhet;
  private int nyhet;
  private int nzhet;
  private int nxnyhet;
  private int nnhet;
  private Double xleft;
  private Double xright;
  private Double yleft;
  private Double yright;
  private Double zleft;
  private Double zright;
  private Double [] xgrid;
  private Double [] ygrid;
  private Double [] zgrid;
  private Double xcenter;
  private Double ycenter;
  private Double zcenter;

  // a thread variable for running the neccesary averaging
  private Thread computeThread;
  private boolean done = false;
  private boolean ok = false;

  // the following integer is a tag to identify the object for use by the
  // object which instantitates one of these averaging things
  private int tag;

}
