/*
 * SimulatorGUI class: FFT
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import java.util.Observable;

/*
 * this simply computes an fft of 2^n complex numbers
 */
public class FFT extends Observable implements Runnable {

  // uniform grid on the unit cubic constructor for the class
  public FFT(double [] d, int isn) throws BadArraySizeException {

    if(testData(d)) data = d;

    // set the tranform / inverse transform parameter
    isign = isn;
    if(isign != 1 && isign != -1) isign = 1;

    // set up the value to return
    value = new double [ndata];

    done = false;

    // set up the thread and run it to compute the average value for the data.
    computeThread = new Thread(this);
    computeThread.start();

  }

  public boolean testData(double [] d) throws BadArraySizeException {
    if(d == null) throw
         new BadArraySizeException("Array size must be even in FFT");
    int n = d.length;
    ndata = n / 2;
    if(2*ndata != n) throw
         new BadArraySizeException("Array size must be even in FFT");
    if(ndata < 1) throw
         new BadArraySizeException("Array the wrong size in FFT");
    int nlog = (int) (Math.log((double) ndata)/Math.log(2.0) + 0.5);
    if(ndata != Math.pow(2,(double)nlog)) throw
         new BadArraySizeException("Array the wrong size in FFT");
    return true;
  }

  public void run() {
    if(computeValue()) {
      ok = true;
    } else {
      ok = false;
    }
    done = true;
    computeThread = null;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for simple averaging                                         *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public double [] computeBitReversal(double [] d) {
    // form the bit reversed array
    int n = d.length;
    int j = 1;
    for(int i=1; i<=n; i=i+2) {
      if(j > i) {
        double tempr = d[j-1];
        double tempi = d[j];
        d[j-1] = d[i-1];
        d[j] = d[i];
        d[i-1] = tempr;
        d[i] = tempi;
      }
      int m = n / 2;
      while((m>=2) && (j>m)) {
        j = j - m;
        m = m / 2;
      }
      j = j + m;
    }
    return d;
  }

  public double [] fft(double [] d) {
    int n = d.length;
    int mmax = 2;
    while(n > mmax) {
      int istep = 2 * mmax;
      double theta = 6.28318530717959 / ((double) isign * mmax );
      double wpr = -2.0 * Math.pow(Math.sin(0.5*theta), 2.0);
      double wpi = Math.sin(theta);
      double wr = 1.0;
      double wi = 0.0;
      for(int m=1; m<=mmax; m=m+2) {
        for(int i=m; i<=n; i=i+istep) {
          int j = i + mmax;
          double tempr = wr * d[j-1] - wi * d[j];
          double tempi = wr * d[j] + wi * d[j-1];
          d[j-1] = d[i-1] - tempr;
          d[j] = d[i] - tempi;
          d[i-1] = d[i-1] + tempr;
          d[i] = d[i] + tempi;
        }
        double wtemp = wr;
        wr = wr * wpr - wi * wpi + wr;
        wi = wi * wpr + wtemp * wpi + wi;
      }
      mmax = istep;
    }
    double fac = Math.sqrt((double) n/2);
    for(int i=0; i<n; i++) d[i] = d[i] / fac;
    return d;
  }

  public boolean computeValue() {
    data = computeBitReversal(data);
    value = fft(data);
    // notify the observers of the object and return
    notifyObservers();
    return true;
  }

  public void setData(double[] d) throws BadArraySizeException {
    if(!testData(d)) return;
    data = d;
    done = false;
    computeThread = new Thread(this);
    computeThread.start();
  }

  public double[] getValues() {
    return value;
  }

  public double[] getAbs() {
    if(value == null) return null;
    abs = new double[ndata];
    for(int i=0; i<ndata; i++)
      abs[i] = value[2*i] * value[2*i] + value[2*i+1] * value[2*i+1];
    return abs;
  }

  public double[] getRe() {
    if(value == null) return null;
    re = new double[ndata];
    for(int i=0; i<ndata; i++) re[i] = value[2*i];
    return re;
  }

  public double[] getIm() {
    if(value == null) return null;
    im = new double[ndata];
    for(int i=0; i<ndata; i++) im[i] = value[2*i+1];
    return im;
  }

  public boolean getStatus() {
    return done;
  }

  public String toString() {
    String s = "";
    return s;
  }
 
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    main method for testing the interface                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public static void main(String args[]) {
    int i, j, k, l;
    int nd = 64;
    // a first simple test on a single block
    FFT aa;
    double [] tdata = new double[2*nd];
    double [] input = new double[2*nd];
    double [] output = new double[2*nd];
    for(i=1; i<2*nd; i++) tdata[i] = 0.0;
    for(i=0; i<nd; i=i+2) tdata[i] = -1.0;
    for(i=0; i<nd; i=i+2) tdata[i+nd] = 1.0;
    input = tdata;
    for(i=0; i<2*nd; i++)
             System.out.println("input:  " + i + " -> " + input[i]);
    try {
      aa = new FFT(input, 1);
      while(!aa.getStatus()) {}
      output = aa.getValues();
      for(i=0; i<2*nd; i++)
               System.out.println("output:  " + i + " -> "+ output[i]);
    } catch(BadArraySizeException base) {
      System.out.println("bad array size caught");
    }
    input = output;
    try {
      aa = new FFT(input, -1);
      while(!aa.getStatus()) {}
      output = aa.getValues();
      for(i=0; i<2*nd; i++)
               System.out.println("output:  " + i + " -> "+ output[i]);
      System.out.println("testing setData()");
      input = tdata;
      aa.setData(input);
      while(!aa.getStatus()) {}
      output = aa.getValues();
      for(i=0; i<2*nd; i++)
               System.out.println("output:  " + i + " -> "+ output[i]);
      double [] abs = aa.getAbs();
      double [] re = aa.getRe();
      double [] im = aa.getIm();
      for(i=0; i<nd; i++) {
        System.out.println("output:  " + i + " -> "+ output[2*i]);
        System.out.println("output:  " + i + " -> "+ output[2*i+1]);
        System.out.println("abs:     " + i + " -> "+ abs[i]);
        System.out.println("re:      " + i + " -> "+ re[i]);
        System.out.println("im:      " + i + " -> "+ im[i]);
      }
    } catch(BadArraySizeException base) {
      System.out.println("bad array size caught");
    }
  }

  // local variables
  private double [] data;
  private int isign = 1;
  private double [] value;
  private double [] abs;
  private double [] im;
  private double [] re;
  private int ndata = 0;

  // a thread variable for running the neccesary averaging
  private Thread computeThread;
  private boolean done = false;
  private boolean ok = false;

  // the following integer is a tag to identify the object for use by the
  // object which instantitates one of these averaging things
  private int tag;

}

