/*
 * SimulatorGUI class: HomogABAverage
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import java.util.Observable;

/*
 * this is an object that will compute the arithmetic average and store the
 * appropriate value for retrieval
 */
public class HomogABAverage extends AverageObject {

  // uniform grid on the unit cubic constructor for the class
  public HomogABAverage(Double [][] d, int nx, int ny, int nz, int t)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         null, null, null);
  }

  // uniform grid on a specified region constructor for the class
  public HomogABAverage(Double [][] d, int nx, int ny, int nz, int t,
                           Double xl, Double xr,
                           Double yl, Double yr,
                           Double zl, Double zr)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         xl, xr, yl, yr, zl, zr,
         null, null, null);
  }

  // specified grid constructor for the class
  public HomogABAverage(Double [][] d, int nx, int ny, int nz, int t,
                           Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         xg, yg, zg);
  }

  // full constructor for the class
  public HomogABAverage(Double [][] d, int nx, int ny, int nz, int t,
                           Double xl, Double xr,
                           Double yl, Double yr,
                           Double zl, Double zr,
                           Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    data = d;
    nxhet = nx;
    nyhet = ny;
    nzhet = nz;
    tag = t;
    nxnyhet = nxhet * nyhet;
    nnhet = nxnyhet * nzhet;
    if(data.length != nnhet) throw
         new BadArraySizeException("Array the wrong size in HomogABAverage");

    // set up the grid
    if(xl != null) {
      xleft = xl;
    } else {
      xleft = new Double(0.0);
    }
    if(xr != null) {
      xright= xr;
    } else {
      xright = new Double(1.0);
    }
    if(yl != null) {
      yleft = yl;
    } else {
      yleft = new Double(0.0);
    }
    if(yr != null) {
      yright= yr;
    } else {
      yright = new Double(1.0);
    }
    if(zl != null) {
      zleft = zl;
    } else {
      zleft = new Double(0.0);
    }
    if(zr != null) {
      zright= zr;
    } else {
      zright = new Double(1.0);
    }

    // make sure the end points are ok
    if(xleft.doubleValue() > xright.doubleValue()) {
      Double temp = xleft;
      xleft = xright;
      xright = temp;
    }
    if(yleft.doubleValue() > yright.doubleValue()) {
      Double temp = yleft;
      yleft = yright;
      yright = temp;
    }
    if(zleft.doubleValue() > zright.doubleValue()) {
      Double temp = zleft;
      zleft = zright;
      zright = temp;
    }

    // set or compute the grid lines in the domain
    if(xg != null) {
      xgrid = xg;
      nxhet = xgrid.length;
    } else {
      xgrid = computeGrid(xleft, xright, nxhet);
    }
    Arrays.sort(xgrid);
    if(yg != null) {
      ygrid = yg;
      nyhet = ygrid.length;
    } else {
      ygrid = computeGrid(yleft, yright, nyhet);
    }
    Arrays.sort(ygrid);
    if(zg != null) {
      zgrid = zg;
      nzhet = zgrid.length;
    } else {
      zgrid = computeGrid(zleft, zright, nzhet);
    }
    Arrays.sort(zgrid);

    // test the end points against the grid - the only time this matters is if
    // the grid lines are specified
    if(xleft.doubleValue() != xgrid[0].doubleValue()) xgrid[0] = xleft;
    if(xright.doubleValue() != xgrid[nxhet].doubleValue())
            xgrid[nxhet] = xright;
    if(yleft.doubleValue() != ygrid[0].doubleValue()) ygrid[0] = yleft;
    if(yright.doubleValue() != ygrid[nyhet].doubleValue())
            ygrid[nyhet] = yright;
    if(zleft.doubleValue() != zgrid[0].doubleValue()) zgrid[0] = zleft;
    if(zright.doubleValue() != zgrid[nzhet].doubleValue())
            zgrid[nzhet] = zright;

    // set up the computational parameters
    done = false;
    ok = false;

    // set up the value to return
    value = new Double[9];

    // set up the thread and run it to compute the average value for the data.
    computeThread = new Thread(this);
    computeThread.start();

  }

  public void run() {
    computeCellCenter();
    if(testData()) {
      computeCellCenter();
      if(computeValue()) {
        ok = true;
      } else {
        ok = false;
      }
    } else {
      ok = false;
    }
    done = true;
    computeThread = null;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for periodic averaging                                       *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public boolean testData() {
    for (int k=1;k<=nzhet;k++) {
      int k2 = ( k - 1 ) * nxnyhet;
      for (int j=1;j<=nyhet;j++) {
        int j2 = ( j - 1 ) * nxhet + k2;
        for (int i=1;i<=nxhet;i++) {
          int index = i + j2 - 1;
          for(int l=0; l<9; l++) {
            if(data[index][l] == null) {
              System.out.println("HomogABAverage: data test failure:"
                                + "\n   x index =       " + i
                                + "\n   y index =       " + j
                                + "\n   z index =       " + k
                                + "\n   spatial index = " + index
                                + "\n   component index = " + l);
              return false;
            }
          }
        }
      }
    }
    return true;
  }

  public boolean computeValue() {
    int j2, k2, index;
    double blksize;
    double totalVolume
               = ( xgrid[nxhet].doubleValue() - xgrid[0].doubleValue() )
               * ( ygrid[nyhet].doubleValue() - ygrid[0].doubleValue() )
               * ( zgrid[nzhet].doubleValue() - zgrid[0].doubleValue() );
    double [] sum = new double[9];
    for(int l=0; l<9; l++) sum[l] = 0.0;
    for (int k=1;k<=nzhet;k++) {
      k2 = ( k - 1 ) * nxnyhet;
      double dz = zgrid[k].doubleValue() - zgrid[k-1].doubleValue();
      for (int j=1;j<=nyhet;j++) {
        j2 = ( j - 1 ) * nxhet + k2;
        double dy = ( ygrid[j].doubleValue() - ygrid[j-1].doubleValue() );
        for (int i=1;i<=nxhet;i++) {
          index = i + j2 - 1;
          double dx = ( xgrid[i].doubleValue() - xgrid[i-1].doubleValue() );
          blksize = dx * dy * dz;
          for(int l=0; l<9; l++) 
                sum[l] = sum[l] + data[index][l].doubleValue() * blksize;
        }
      }
    }
    for(int l=0; l<9; l++) value[l] = new Double(sum[l] / totalVolume);
    notifyObservers();
    return true;
  }

  public Double [] computeGrid(Double xl, Double xr, int nx) {
    Double [] grid = null;
    grid = new Double [nx+1];
    double dx = (xr.doubleValue() - xl.doubleValue()) / ((double) nx);
    grid[0] = xl;
    for(int i=1; i<nx; i++) {
      grid[i] = new Double(grid[i-1].doubleValue() + dx);
    }
    grid[nx] = xr;
    return grid;
  }

  public void computeCellCenter() {
    xcenter = new Double(0.5 * ( xleft.doubleValue() + xright.doubleValue() ));
    ycenter = new Double(0.5 * ( yleft.doubleValue() + yright.doubleValue() ));
    zcenter = new Double(0.5 * ( zleft.doubleValue() + zright.doubleValue() ));
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for getting the computed value at an intermediate point      *
   *                                                                         *
   *     note that a capital "F" indicates a return type of Double while a   *
   *          small "f' indicates a return type of double                    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getFunctionValue(Double x, Double y, Double z) {
    return getFunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public Double [] getFunctionValue(double x, double y, double z) {
    return value;
  }

  public double [] getfunctionValue(Double x, Double y, Double z) {
    return getfunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public double [] getfunctionValue(double x, double y, double z) {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the computed value                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getDoubleValue() {
    return value;
  }

  public double [] getdoubleValue() {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the internal variables - the most important is    *
   * probably getTag which returns the integer identification of the object  *
   * which must be set in the instantiation                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
 
  public Double [][] getData() {
    return data;
  }
 
  public int getNumberOfXBlocks() {
    return nxhet;
  }
 
  public int getNumberOfYBlocks() {
    return nyhet;
  }
 
  public int getNumberOfZBlocks() {
    return nzhet;
  }
 
  public int getTotalBlocks() {
    return nnhet;
  }

  public Double getXCenter() {
    return xcenter;
  }

  public Double getYCenter() {
    return ycenter;
  }

  public Double getZCenter() {
    return zcenter;
  }

  public double getxCenter() {
    return xcenter.doubleValue();
  }

  public double getyCenter() {
    return ycenter.doubleValue();
  }

  public double getzCenter() {
    return zcenter.doubleValue();
  }

  public Double getXLeft() {
    return xleft;
  }

  public double getxLeft() {
    return xleft.doubleValue();
  }

  public Double getXRight() {
    return xright;
  }

  public double getxRight() {
    return xright.doubleValue();
  }

  public Double getYLeft() {
    return yleft;
  }

  public double getyLeft() {
    return yleft.doubleValue();
  }

  public Double getYRight() {
    return yright;
  }

  public double getyRight() {
    return yright.doubleValue();
  }

  public Double getZLeft() {
    return zleft;
  }

  public double getzLeft() {
    return zleft.doubleValue();
  }

  public Double getZRight() {
    return zright;
  }

  public double getzRight() {
    return zright.doubleValue();
  }
 
  public int getTag() {
    return tag;
  }
 
  public boolean isValueOK() {
    return ok;
  }
 
  public boolean getStatus() {
    return done;
  }
 
  public String toString() {
    String s = "";
    s = "\nString output for the HomogAB Average Object\n\n";
    s = s + " |  ";
    if(value[0] != null) {
      s = s + "axx= " + value[0].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[1] != null) {
      s = s + "axy= " + value[1].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[2] != null) {
      s = s + "axz= " + value[2].toString();
    } else {
      s = s + "null";
    }
    s = s + "  |\n";
    s = s + " |  ";
    if(value[3] != null) {
      s = s + "ayx= " + value[3].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[4] != null) {
      s = s + "ayy= " + value[4].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[5] != null) {
      s = s + "ayz= " + value[5].toString();
    } else {
      s = s + "null";
    }
    s = s + "  |\n";
    s = s + " |  ";
    if(value[6] != null) {
      s = s + "azx= " + value[6].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[7] != null) {
      s = s + "azy= " + value[7].toString();
    } else {
      s = s + "null";
    }
    s = s + "   ";
    if(value[8] != null) {
      s = s + "azz= " + value[8].toString();
    } else {
      s = s + "null";
    }
    s = s + "  |\n";
    return s;
  }
 
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    main method for testing the interface                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public static void main(String args[]) {
    int i, j, k, l;
    int nx = 10;
    int ny = 10;
    int nz = 1;
    int ntot = nx * ny * nz;
    // a first simple test on a single block
    HomogABAverage aa;
    Double [][] data = new Double[100][9];
    for(i=0; i<100; i++) for(l=0; l<9; l++) data[i][l] = new Double(i);
    try {
      aa = new HomogABAverage(data, nx, ny, nz, 1);
      System.out.println("getData() test:   ");
      data = aa.getData();
      for(l=0; l<9; l++) {
        for(i=0; i<ntot; i++) {
           System.out.println("Data Value: " + i + " " + data[i][l]);
        }
      }
      while(!aa.getStatus()) {
        System.out.println("isValueOK() test:   " + aa.isValueOK());
        System.out.println("getStatus() test:   " + aa.getStatus());
        System.out.println("   ");
        // sleep the process and test again
        try {Thread.sleep(1000);
        } catch(InterruptedException ie) {
          break;
        }
      }
      System.out.println("Double Output: " + aa.toString());
      System.out.println("double Output: " + aa.getdoubleValue());
      System.out.println("isValueOK() test:   " + aa.isValueOK());
      System.out.println("getStatus() test:   " + aa.getStatus());
      nx = aa.getNumberOfXBlocks();
      System.out.println("getNumberOfXBlocks() test:   " + nx);
      ny = aa.getNumberOfYBlocks();
      System.out.println("getNumberOfYBlocks() test:   " + ny);
      nz = aa.getNumberOfZBlocks();
      System.out.println("getNumberOfZBlocks() test:   " + nz);
      System.out.println("getXLeft() test:             " + aa.getXLeft());
      System.out.println("getxLeft() test:             " + aa.getxLeft());
      System.out.println("getXRight() test:            " + aa.getXRight());
      System.out.println("getxRight() test:            " + aa.getxRight());
      System.out.println("getYLeft() test:             " + aa.getYLeft());
      System.out.println("getyLeft() test:             " + aa.getyLeft());
      System.out.println("getYRight() test:            " + aa.getYRight());
      System.out.println("getyRight() test:            " + aa.getyRight());
      System.out.println("getZLeft() test:             " + aa.getZLeft());
      System.out.println("getzLeft() test:             " + aa.getzLeft());
      System.out.println("getZRight() test:            " + aa.getZRight());
      System.out.println("getzRight() test:            " + aa.getzRight());
      System.out.println("getXCenter() test:           " + aa.getXCenter());
      System.out.println("getYCenter() test:           " + aa.getYCenter());
      System.out.println("getZCenter() test:           " + aa.getZCenter());
      System.out.println("getxCenter() test:           " + aa.getxCenter());
      System.out.println("getyCenter() test:           " + aa.getyCenter());
      System.out.println("getzCenter() test:           " + aa.getzCenter());
      ntot = aa.getTotalBlocks();
      System.out.println("getTotalBlocks() test:   " + ntot);
      System.out.println("getTag() test:   " + aa.getTag());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }
  }

  // local variables
  private Double [] value;
  private Double [][] data;
  private int nxhet;
  private int nyhet;
  private int nzhet;
  private int nxnyhet;
  private int nnhet;
  private Double xleft;
  private Double xright;
  private Double yleft;
  private Double yright;
  private Double zleft;
  private Double zright;
  private Double [] xgrid;
  private Double [] ygrid;
  private Double [] zgrid;
  private Double xcenter;
  private Double ycenter;
  private Double zcenter;

  // a thread variable for running the neccesary averaging
  private Thread computeThread;
  private boolean done = false;
  private boolean ok = false;

  // the following integer is a tag to identify the object for use by the
  // object which instantitates one of these averaging things
  private int tag;

}

