import java.io.*;
import java.util.StringTokenizer;
/*
 */

public class Mesh1D extends Object {

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  Constructors for the class                                             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  // constructor with no arguments - all defaults
  public Mesh1D() {
    this(0.0, 1.0, null, -1, null, -1, null);
  }

  public Mesh1D(double xl, double xr, int nb) {
    this(xl, xr, null, -1, null, nb, null);
  }

  public Mesh1D(double [] xgr) {
    this(0.0, 1.0, xgr, -1, null, -1, null);
  }

  public Mesh1D(double xl, double xr, double [] xn) {
    this(xl, xr, null, -1, xn, -1, null);
  }

  // constructor with all arguments
  public Mesh1D(double xl,
                double xr,
                double [] xgr,
                int ng,
                double [] xn,
                int nn,
                String f) {

    // set the boundary values in the correct order
    xleft = xl;
    xright = xr;
    if(xleft > xright) {
      xleft = xr;
      xright = xl;
    }

    // test for a uniform grid - if the arrays of grid locations and nodes are
    // null the code assumes the grid is uniform
    if(xgr == null && xn == null) {

      // uniform grid based on the boundaries
      nxgr = nn + 1;
      if(nxgr < 2) nxgr = DEFAULT_NUMBER_OF_BLOCKS + 1;
      nx = nxgr - 1;
      double delx = ( xright - xleft ) / (double) nx;
      xgrid = new double [nxgr];
      for(int i=0; i<nxgr; i++) {
        xgrid[i] = xleft + ((double) i) * delx;
      }
      computeNodes();
      computeBlockSizes();

    } else {

      // nonuniform cases
      if(xgr != null) {
        xgrid = sort(xgr);
        nxgr = xgrid.length;
        xleft = xgrid[0];
        xright = xgrid[nxgr-1];
        computeNodes();
        computeBlockSizes();
      } else {
        xnode = sort(xn);
        nx = xn.length;
        if(xleft > xnode[0]) xleft = 1.5 * xnode[0] - 0.5 * xnode[0];
        if(xright < xnode[nx-1]) xright = 1.5 * xnode[nx-1] - 0.5 * xnode[nx-2];
        nx = xnode.length;
        computeGridLines();
        computeBlockSizes();
      }

    }

    // set the filename string
    filename = f;

  }

  public void computeNodes() {
    // make sure the grid line values are available to do the calculations
    if(xgrid == null) return;
    // make sure there are enough nodes to do the calculations
    if(nxgr < 2) return;
    nx = nxgr - 1;
    xnode = new double [nx];
    if(xnode == null) xnode = new double [nx];
    for(int i=0; i<nx; i++) {
      xnode[i] = 0.5 * ( xgrid[i] + xgrid[i+1] );
    }
  }

  public void computeGridLines() {
    // make sure the nodes are available to do the calculations
    if(xnode == null) return;
    // make sure there are enough nodes to do the calculations
    if(nx < 2) return;
    nxgr = nx + 1;
    xgrid = new double [nxgr];
    xgrid[0] = xleft;
    for(int i=1; i<nx; i++) {
      xgrid[i] = 0.5 * ( xnode[i-1] + xnode[i] );
    }
    xgrid[nx] = xright;
  }

  public void computeBlockSizes() {
    if(xgrid != null) {
      nx = nxgr - 1;
      dx = new double [nx];
      for(int i=0; i<nx; i++) dx[i] = ( xgrid[i+1] - xgrid[i] );
      for(int i=0; i<nx; i++) {
        if(meshSize < dx[i]) meshSize = dx[i];
      }
      return;
    }
    if(xnode != null) {
      computeGridLines();
      dx = new double [nx];
      for(int i=0; i<nx; i++) dx[i] = ( xgrid[i+1] - xgrid[i] );
      for(int i=0; i<nx; i++) {
        if(meshSize < dx[i]) meshSize = dx[i];
      }
      return;
    }
  }

  public int getIndex(double x) {
    if(xgrid != null) {
      for(int i=0; i<nx; i++) {
        if((xgrid[i] <= x) && (x < xgrid[i+1])) return i;
      }
    }
    return -1;
  }

  public double [] getGridLines() {
    return xgrid;
  }

  public int getNumberOfGridLines() {
    return nxgr;
  }

  public double [] getNodes() {
    return xnode;
  }

  public double [] getGridBlockSizes() {
    return dx;
  }

  public int getNumberOfNodes() {
    return nx;
  }

  public double getMeshSize() {
    return meshSize;
  }

  public double getLeftBoundary() {
    return xleft;
  }

  public double getRightBoundary() {
    return xright;
  }

  public void addNode(int ind) {
    if(xnode == null) return;
    int n = getNumberOfNodes();
    double [] temp = new double [n+1];
    if(ind < 1) {
      temp[0] = 1.5 * xnode[0] - 0.5 * xnode[1];
      for(int i=1; i<n+1; i++) temp[i] = xnode[i-1];
    }
    if(ind > 0 && ind < n-1) {
      for(int i=0; i<ind; i++) temp[i] = xnode[i];
      temp[ind] = 0.5 * ( xnode[ind-1] + xnode[ind] );
      for(int i=ind+1; i<n+1; i++) temp[i] = xnode[i-1];
    }
    if(ind >= n) {
      for(int i=0; i<n; i++) temp[i] = xnode[i-1];
      temp[n] = 1.5 * xnode[nx-2] - 0.5 * xnode[nx-1];
    }
    nx = nx + 1;
    xnode = sort(temp);
    computeGridLines();
    computeBlockSizes();
  }

  public void addNode(double x) {
    if(xnode == null) return;
    int n = getNumberOfNodes();
    double [] temp = new double [n+1];
    for(int i=0; i<nx; i++) temp[i] = xnode[i];
    temp[nx] = x;
    nx = nx + 1;
    xnode = sort(temp);
    computeGridLines();
    computeBlockSizes();
  }

  public void removeNode(int ind) {
    if(xnode == null) return;
    int n = getNumberOfNodes();
    double [] temp = new double [n-1];
    if(ind < 1) {
      for(int i=0; i<n-1; i++) temp[i] = xnode[i+1];
    }
    if(ind > 0 && ind < n-1) {
      for(int i=0; i<ind; i++) temp[i] = xnode[i];
      for(int i=ind+1; i<n; i++) temp[i-1] = xnode[i];
    }
    if(ind >= n) {
      for(int i=0; i<n-1; i++) temp[i] = xnode[i];
    }
    nx = nx - 1;
    xnode = sort(temp);
    computeGridLines();
    computeBlockSizes();
  }

  public void removeNode(double x) {
    int ind = getIndex(x);
    removeNode(ind);
  }

  public void refineBy2() {
    if(xgrid == null) return;
    int n = getNumberOfNodes();
    double [] temp = new double [2*n+1];
    for(int i=0; i<n; i++) {
      temp[2*i] = xgrid[i];
      temp[2*i+1] = 0.5 * ( xgrid[i] + xgrid[i+1] );
    }
    temp[2*n] = xgrid[n];
    xgrid = temp;
    nxgr = 2 * n + 1;
    nx = nxgr - 1;
    computeNodes();
    computeBlockSizes();
  }

  public void coarsenBy2() {
    if(xgrid == null) return;
    int n = getNumberOfNodes();
    if(n<4) return;
    double [] temp = new double [n/2+1];
    for(int i=0; i<n/2; i++) {
      temp[i] = xgrid[2*i];
    }
    temp[n/2] = xgrid[nx];
    nx = n / 2;
    xgrid = temp;
    nxgr = nx + 1;
    computeNodes();
    computeBlockSizes();
  }

  public String toString() {
    String s = "";
    s = "\n Mesh1D Object Information\n";
    s = s + "\n left boundary endpoint = " + digitFilter(xleft,5);
    s = s + "\n right boundary endpoint = " + digitFilter(xright,5);
    s = s + "\n number of grid lines = " + new Integer(nxgr).toString();
    for(int i=0; i<nxgr; i++)
         s = s + "\n grid line i = " + i + " -> " + digitFilter(xgrid[i],5);
    s = s + "\n number of nodes = " + new Integer(nx).toString();
    for(int i=0; i<nx; i++)
         s = s + "\n node i = " + i + " -> " + digitFilter(xnode[i],5);
    s = s + "\n number of blocks = " + new Integer(nx).toString();
    for(int i=0; i<nx; i++)
         s = s + "\n block size i = " + i + " -> " + digitFilter(dx[i],5);
    s = s + "\n";
    return s;
  }

  private String digitFilter(double value, int n) {
    String integerPart = "0";
    String dot = ".";
    String fractionPart = "0";
    if(Math.abs(value) < .000000000001) value = 0.0;
    String s = Double.toString(value);
    StringTokenizer st = new StringTokenizer(s, " .");
    if(st.hasMoreTokens()) integerPart = st.nextToken();
    if(st.hasMoreTokens()) {
      String t = st.nextToken();
      if(t.length() <= n) fractionPart = t;
      if(t.length() > n) fractionPart = t.substring(0,n);
    }
    return integerPart + dot + fractionPart;
  }

  public double [] sort(double[] input) {
    int n = input.length;
    int i;
    for(int j=1; j<n; j++) {
      double s = input[j];
      for(i=j-1; i>=0; i--) {
        if(s > input[i]) break;
        input[i+1] = input[i];
      }
      input[i+1] = s;
    }
    return input;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   * main() - called when the application is run.                            *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public static void main(String args[]) {
    System.out.println(" *************** Test #1 *******************");
    System.out.println(" Uniform Grid Test ");
    Mesh1D m1d1 = new Mesh1D();
    System.out.println(m1d1.toString());
    System.out.println(" *************** Test #2 *******************");
    System.out.println(" Nonuniform Grid Test - Grid Input");
    double [] xgr = new double [13];
    for(int i=0; i<13; i++) xgr[i] = Math.random();
    Mesh1D m1d2 = new Mesh1D(xgr);
    System.out.println(m1d2.toString());
    System.out.println(" *************** Test #3 *******************");
    System.out.println(" Nonuniform Grid Test - Node Input");
    double [] xn = new double [17];
    for(int i=0; i<17; i++) xn[i] = Math.random();
    Mesh1D m1d3 = new Mesh1D(0.0, 1.0, xn);
    System.out.println(m1d3.toString());
    System.out.println(" *************** Test #4 *******************");
    System.out.println(" Nonuniform Grid Test - Node Input");
    xn = new double [11];
    for(int i=0; i<11; i++) xn[i] = Math.random();
    Mesh1D m1d4 = new Mesh1D(0.5, 0.6, xn);
    System.out.println(m1d4.toString());
    System.out.println(" *************** Test #5 *******************");
    System.out.println(" Nonuniform Grid Test - Add a Node");
    m1d4.addNode(Math.random());
    System.out.println(m1d4.toString());
    System.out.println(" *************** Test #6 *******************");
    System.out.println(" Nonuniform Grid Test - Add a Node");
    m1d4.addNode(7);
    System.out.println(m1d4.toString());
    System.out.println(" *************** Test #7 *******************");
    System.out.println(" Nonuniform Grid Test - Remove a Node");
    m1d4.removeNode(7);
    System.out.println(m1d4.toString());
    System.out.println(" *************** Test #8 *******************");
    System.out.println(" Nonuniform Grid Test - Remove a Node");
    m1d4.removeNode(0.5);
    System.out.println(m1d4.toString());
    System.out.println(" *************** Test #9 *******************");
    System.out.println(" Uniform Grid Test - Refine by 2 ");
    m1d1.refineBy2();
    System.out.println(m1d1.toString());
    System.out.println(" *************** Test #10 *******************");
    System.out.println(" Uniform Grid Test - Coarsen by 2 ");
    m1d1.coarsenBy2();
    System.out.println(m1d1.toString());
  }

  private double xleft;
  private double xright;
  private double [] xgrid;
  private int nxgr = 0;
  private double [] xnode;
  private int nx = 0;
  private double [] dx;

  private double meshSize = 0.0;

  private int DEFAULT_NUMBER_OF_BLOCKS = 5;

  private String filename;
  private File file;

}
