/*
 * SimulatorGUI class: PeriodicWaveletAverage2DNew
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Arrays;
import java.util.Observable;
import java.util.StringTokenizer;

/*
 * this is an object that will compute the apreiodic average and store the
 * appropriate value for retrieval
 */
public class PeriodicWaveletAverage2DNew extends AverageObject {

  // uniform grid on the unit cubic constructor for the class
  public PeriodicWaveletAverage2DNew(Double[][] d, int nx, int ny, int nz,
                                     int t)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         null, null, null);
  }

  // uniform grid on a specified region constructor for the class
  public PeriodicWaveletAverage2DNew(Double[][] d, int nx, int ny, int nz,
                                     int t,
                                     Double xl, Double xr,
                                     Double yl, Double yr,
                                     Double zl, Double zr)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         xl, xr, yl, yr, zl, zr,
         null, null, null);
  }

  // specified grid constructor for the class
  public PeriodicWaveletAverage2DNew(Double[][] d, int nx, int ny, int nz,
                                     int t,
                                     Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    this(d, nx, ny, nz, t,
         null, null, null, null, null, null,
         xg, yg, zg);
  }

  // full constructor for the class
  public PeriodicWaveletAverage2DNew(Double[][] d, int nx, int ny, int nz,
                                     int t,
                                     Double xl, Double xr,
                                     Double yl, Double yr,
                                     Double zl, Double zr,
                                     Double [] xg, Double [] yg, Double [] zg)
       throws BadArraySizeException {
    data = d;
    nxhet = nx;
    nyhet = ny;
    nzhet = nz;
    tag = t;
    nxnyhet = nxhet * nyhet;
    nnhet = nxnyhet * nzhet;
    if(data.length != nnhet) throw
       new BadArraySizeException("Bad Array size: PeriodicWaveletAverage2DNew");

    // set up the grid
    if(xl != null) {
      xleft = xl;
    } else {
      xleft = new Double(0.0);
    }
    if(xr != null) {
      xright= xr;
    } else {
      xright = new Double(1.0);
    }
    if(yl != null) {
      yleft = yl;
    } else {
      yleft = new Double(0.0);
    }
    if(yr != null) {
      yright= yr;
    } else {
      yright = new Double(1.0);
    }
    if(zl != null) {
      zleft = zl;
    } else {
      zleft = new Double(0.0);
    }
    if(zr != null) {
      zright= zr;
    } else {
      zright = new Double(1.0);
    }

    // make sure the end points are ok
    if(xleft.doubleValue() > xright.doubleValue()) {
      Double temp = xleft;
      xleft = xright;
      xright = temp;
    }
    if(yleft.doubleValue() > yright.doubleValue()) {
      Double temp = yleft;
      yleft = yright;
      yright = temp;
    }
    if(zleft.doubleValue() > zright.doubleValue()) {
      Double temp = zleft;
      zleft = zright;
      zright = temp;
    }

    // set or compute the grid lines in the domain
    if(xg != null) {
      xgrid = xg;
      nxhet = xgrid.length;
    } else {
      xgrid = computeGrid(xleft, xright, nxhet);
    }
    Arrays.sort(xgrid);
    if(yg != null) {
      ygrid = yg;
      nyhet = ygrid.length;
    } else {
      ygrid = computeGrid(yleft, yright, nyhet);
    }
    Arrays.sort(ygrid);
    if(zg != null) {
      zgrid = zg;
      nzhet = zgrid.length;
    } else {
      zgrid = computeGrid(zleft, zright, nzhet);
    }
    Arrays.sort(zgrid);

    // test the end points against the grid - the only time this matters is if
    // the grid lines are specified
    if(xleft.doubleValue() != xgrid[0].doubleValue()) xgrid[0] = xleft;
    if(xright.doubleValue() != xgrid[nxhet].doubleValue())
            xgrid[nxhet] = xright;
    if(yleft.doubleValue() != ygrid[0].doubleValue()) ygrid[0] = yleft;
    if(yright.doubleValue() != ygrid[nyhet].doubleValue())
            ygrid[nyhet] = yright;
    if(zleft.doubleValue() != zgrid[0].doubleValue()) zgrid[0] = zleft;
    if(zright.doubleValue() != zgrid[nzhet].doubleValue())
            zgrid[nzhet] = zright;

    // set up the computational parameters
    done = false;
    ok = false;

    // set up the value to return
    value = new Double[9];

    // set up the thread and run it to compute the average value for the data.
    computeThread = new Thread(this);
    computeThread.start();

  }

  public void run() {
    computeCellCenter();
    if(testData()) {
      computeCellCenter();
      if(computeValue()) {
        ok = true;
      } else {
        ok = false;
      }
    } else {
      ok = false;
    }
    done = true;
    computeThread = null;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for simple averaging                                         *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public boolean testData() {
    for (int k=1;k<=nzhet;k++) {
      int k2 = ( k - 1 ) * nxnyhet;
      for (int j=1;j<=nyhet;j++) {
        int j2 = ( j - 1 ) * nxhet + k2;
        for (int i=1;i<=nxhet;i++) {
          int index = i + j2 - 1;
          for(int l=0; l<9; l++) {
            if(data[index][l] == null) {
            System.out.println("PeriodicWaveletAverage2DNew: data test failure:"
                               + "\n   x index =       " + i
                               + "\n   y index =       " + j
                               + "\n   z index =       " + k
                               + "\n   spatial index = " + index
                               + "\n   component index = " + l);
              return false;
            }
          }
        }
      }
    }
    return true;
  }

  public boolean computeValue() {

    // test for an appropriate grid size - some power of two in each coordinate
    // direction
    if(!testNumberOfBlocks()) return false;

    // at this point we have a dyadic grid and can proceed with the calculation
    // loop over the levels
    int limVal = nxhet / 2; 
    int skipBy2 = 2;
    for(int l=0; l<powOfTwo; l++) {
      for(int j=0; j<limVal; j++) {
        int j2 = skipBy2 * j * nxhet; 
        for(int i=0; i<limVal; i++) {
          int index = skipBy2 * i + j2;
	  // pull off the four tensor values for the 2x2 block currently being
	  // considered
          double [][] tensI = setTensor(index);
          double [][] tensII = setTensor(index+skipBy2/2);
          double [][] tensIII = setTensor(index+skipBy2*nxhet/2);
          double [][] tensIV = setTensor(index+skipBy2/2+skipBy2*nxhet/2);

	  // compute some parameters that we need to fill out the solution

          // combinations of the values
          // averages
          double kxxbar = tensI[0][0] + tensII[0][0] + tensIII[0][0]
                        + tensIV[0][0];
          double kxybar = tensI[0][1] + tensII[0][1] + tensIII[0][1]
                        + tensIV[0][1];
          double kyxbar = tensI[0][1] + tensII[0][1] + tensIII[0][1]
                        + tensIV[0][1];
          double kyybar = tensI[1][1] + tensII[1][1] + tensIII[1][1]
                        + tensIV[1][1];
          // tilde's
          double kxxtld = tensI[0][0] - tensII[0][0] - tensIII[0][0]
                        + tensIV[0][0];
          double kxytld = tensI[0][1] - tensII[0][1] - tensIII[0][1]
                        + tensIV[0][1];
          double kyxtld = tensI[0][1] - tensII[0][1] - tensIII[0][1]
                        + tensIV[0][1];
          double kyytld = tensI[1][1] - tensII[1][1] - tensIII[1][1]
                        + tensIV[1][1];
          // hat's
          double kxxhat = tensI[0][0] - tensII[0][0] + tensIII[0][0]
                        - tensIV[0][0];
          double kxyhat = tensI[0][1] - tensII[0][1] + tensIII[0][1]
                        - tensIV[0][1];
          double kyxhat = tensI[0][1] - tensII[0][1] + tensIII[0][1]
                        - tensIV[0][1];
          double kyyhat = tensI[1][1] - tensII[1][1] + tensIII[1][1]
                        - tensIV[1][1];
          // dot's
          double kxxdot = tensI[0][0] + tensII[0][0] - tensIII[0][0]
                        - tensIV[0][0];
          double kxydot = tensI[0][1] + tensII[0][1] - tensIII[0][1]
                        - tensIV[0][1];
          double kyxdot = tensI[0][1] + tensII[0][1] - tensIII[0][1]
                        - tensIV[0][1];
          double kyydot = tensI[1][1] + tensII[1][1] - tensIII[1][1]
                        - tensIV[1][1];

          // compute the coefficients for the scaling function
          double [] output = localCoefficients(kxxbar, kxxhat, kxxtld, kxxdot,
                                               kxybar, kxyhat, kxytld, kxydot,
                                               kyxbar, kyxhat, kyxtld, kyxdot,
                                               kyybar, kyyhat, kyytld, kyydot);

          // compute the common denominator in the solution of the perturbation
          // terms in the homogeniztion
          double denom = 4.0 * ( kxytld * kyxtld - kxxbar * kyybar );

	  // set up the tensor that results from the local 2x2 problem
          double [][] kout = new double[3][3];
          kout[0][0] = 0.25 * kxxbar
                     - ( kxytld * kyxdot - kxxhat * kyybar ) * kxxhat / denom
                     - ( kxxhat * kxytld - kxxbar * kxydot ) * kxydot / denom;
          kout[0][1] = 0.25 * kxybar
                     + ( kxyhat * kyybar - kxytld * kyydot ) * kxxhat / denom
                     - ( kxyhat * kxytld - kxxbar * kyydot ) * kxydot / denom;
          kout[0][2] = 0.0;
          kout[1][0] = kout[0][1];
          kout[1][1] = 0.25 * kyybar
                     + ( kxyhat * kyybar - kxytld * kyydot ) * kyxhat / denom
                     - ( kxyhat * kxytld - kxxbar * kyydot ) * kyydot / denom;
          kout[1][2] = 0.0;
          kout[2][0] = 0.0;
          kout[2][1] = 0.0;
          kout[2][2] = 1.0;

          // reset the appropriate cell to hold the average value at this
	  // level
          getTensor(index, kout);

        }
      }
      limVal = limVal / 2;
      skipBy2 = 2 * skipBy2;
    }

    for(int l=0; l<9; l++) value[l] = data[0][l];

    notifyObservers();
    return true;
  }

  public double [] localCoefficients(double kxxbar, double kxxhat,
                                     double kxxtld, double kxxdot,
                                     double kxybar, double kxyhat,
                                     double kxytld, double kxydot,
                                     double kyxbar, double kyxhat,
                                     double kyxtld, double kyxdot,
                                     double kyybar, double kyyhat,
                                     double kyytld, double kyydot) {

    // set up the right hand side for the two systems of equations
    double [][] rhs = new double[4][2];
    rhs[0][0] = kxxhat;
    rhs[1][0] = kxxdot;
    rhs[2][0] = kxydot;
    rhs[3][0] = kxytld;
    rhs[0][1] = kxyhat;
    rhs[1][1] = kxydot;
    rhs[2][1] = kyydot;
    rhs[3][1] = kyytld;

    // the following produces the coefficients in the transform given the
    // values in the right hand side for the 2x2 case.
    double cdet =  kyybar * kyybar - kyytld * kyytld;
    double d11 = ( kxyhat * kyybar - kxydot * kyytld ) / cdet;
    double d12 = ( -kxyhat * kyytld + kxydot * kyybar ) / cdet;
    double d21 = ( kxybar * kyybar - kxytld * kyytld ) / cdet;
    double d22 = ( -kxybar * kyytld + kxytld * kyybar ) / cdet;
    double s21 = d11 * rhs[2][0] + d12 * rhs[3][0];
    double s22 = d21 * rhs[2][0] + d22 * rhs[3][0];
    double e11 = kxxbar - d11 * kxyhat - d12 * kxydot;
    double e12 = kxxhat - d11 * kxybar - d12 * kxytld;
    double e21 = kxxhat - d21 * kxyhat - d22 * kxydot;
    double e22 = kxxbar - d21 * kxybar - d22 * kxytld;
    double s11 = rhs[0][0] - s21;
    double s12 = rhs[1][0] - s22;
    double edet = e11 * e22 - e21 * e12;
    double x11 = ( e22 * s11 - e12 * s12 ) / edet;
    double x12 = ( - e21 * s11 + e11 * s12 ) / edet;
    double t11 = rhs[2][0] - kxyhat * x11 - kxybar * x12;
    double t12 = rhs[3][0] - kxydot * x11 - kxytld * x12;
    double x21 = ( kyybar * t11 - kyytld * t12 ) / cdet;
    double x22 = ( - kyytld * t11 + kyybar * t12 ) / cdet;
    double [] output = new double [4];
    output[0] = x11;
    output[1] = x12;
    output[2] = x21;
    output[3] = x22;
    return output;

  }

  private double [][] setTensor(int index) {
    if(index < 0 || index >= nnhet) return null;
    double [][] tens = new double[3][3];
    tens[0][0] = data[index][0].doubleValue();
    tens[0][1] = data[index][1].doubleValue();
    tens[0][2] = data[index][2].doubleValue();
    tens[1][0] = data[index][3].doubleValue();
    tens[1][1] = data[index][4].doubleValue();
    tens[1][2] = data[index][5].doubleValue();
    tens[2][0] = data[index][6].doubleValue();
    tens[2][1] = data[index][7].doubleValue();
    tens[2][2] = data[index][8].doubleValue();
    return tens;
  }

  private void getTensor(int index, double [][] tens) {
    if(index < 0 || index >= nnhet) return;
    data[index][0] = new Double(tens[0][0]);
    data[index][1] = new Double(tens[0][1]);
    data[index][2] = new Double(tens[0][2]);
    data[index][3] = new Double(tens[1][0]);
    data[index][4] = new Double(tens[1][1]);
    data[index][5] = new Double(tens[1][2]);
    data[index][6] = new Double(tens[2][0]);
    data[index][7] = new Double(tens[2][1]);
    data[index][8] = new Double(tens[2][2]);
  }

  public boolean testNumberOfBlocks() {
    if(nzhet > 1) return false;
    if(nxhet != nyhet) return false;
    int ival = nxhet;
    int pow2 = 1;
    powOfTwo = 0;
    while(ival>1) {
      ival = ival / 2;
      pow2 = 2 * pow2;
      powOfTwo++;
    }
    return true;
  }

  public Double [] computeGrid(Double xl, Double xr, int nx) {
    Double [] grid = null;
    grid = new Double [nx+1];
    double dx = (xr.doubleValue() - xl.doubleValue()) / ((double) nx);
    grid[0] = xl;
    for(int i=1; i<nx; i++) {
      grid[i] = new Double(grid[i-1].doubleValue() + dx);
    }
    grid[nx] = xr;
    return grid;
  }

  public void computeCellCenter() {
    xcenter = new Double(0.5 * ( xleft.doubleValue() + xright.doubleValue() ));
    ycenter = new Double(0.5 * ( yleft.doubleValue() + yright.doubleValue() ));
    zcenter = new Double(0.5 * ( zleft.doubleValue() + zright.doubleValue() ));
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    methods for getting the computed value at an intermediate point      *
   *                                                                         *
   *     note that a capital "F" indicates a return type of Double while a   *
   *          small "f' indicates a return type of double                    *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getFunctionValue(Double x, Double y, Double z) {
    return getFunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public Double [] getFunctionValue(double x, double y, double z) {
    return value;
  }

  public double [] getfunctionValue(Double x, Double y, Double z) {
    return getfunctionValue(x.doubleValue(), y.doubleValue(), z.doubleValue());
  }

  public double [] getfunctionValue(double x, double y, double z) {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the computed value                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  public Double [] getDoubleValue() {
    return value;
  }

  public double [] getdoubleValue() {
    double [] dval = new double[9]; 
    for(int l=0; l<9; l++) dval[l] = value[l].doubleValue();
    return dval;
  }

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    method for getting the internal variables - the most important is    *
   * probably getTag which returns the integer identification of the object  *
   * which must be set in the instantiation                                  *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
 
  public Double [][] getData() {
    return data;
  }
 
  public int getNumberOfXBlocks() {
    return nxhet;
  }
 
  public int getNumberOfYBlocks() {
    return nyhet;
  }
 
  public int getNumberOfZBlocks() {
    return nzhet;
  }
 
  public int getTotalBlocks() {
    return nnhet;
  }

  public Double getXCenter() {
    return xcenter;
  }

  public Double getYCenter() {
    return ycenter;
  }

  public Double getZCenter() {
    return zcenter;
  }

  public double getxCenter() {
    return xcenter.doubleValue();
  }

  public double getyCenter() {
    return ycenter.doubleValue();
  }

  public double getzCenter() {
    return zcenter.doubleValue();
  }

  public Double getXLeft() {
    return xleft;
  }

  public double getxLeft() {
    return xleft.doubleValue();
  }

  public Double getXRight() {
    return xright;
  }

  public double getxRight() {
    return xright.doubleValue();
  }

  public Double getYLeft() {
    return yleft;
  }

  public double getyLeft() {
    return yleft.doubleValue();
  }

  public Double getYRight() {
    return yright;
  }

  public double getyRight() {
    return yright.doubleValue();
  }

  public Double getZLeft() {
    return zleft;
  }

  public double getzLeft() {
    return zleft.doubleValue();
  }

  public Double getZRight() {
    return zright;
  }

  public double getzRight() {
    return zright.doubleValue();
  }
 
  public int getTag() {
    return tag;
  }
 
  public boolean isValueOK() {
    return ok;
  }
 
  public boolean getStatus() {
    return done;
  }

  public static String digitFilter(Double value, int n) {
    double dval = value.doubleValue();
    return digitFilter(dval, n);
  }

  public static String digitFilter(double value, int n) {
    String integerPart = "0";
    String dot = ".";
    String fractionPart = "0";
    String exponent = "";
    String t = null;
    String s = Double.toString(value);
    StringTokenizer st = new StringTokenizer(s, " .eE");
    if(st.hasMoreTokens()) integerPart = st.nextToken();
    if(st.hasMoreTokens()) t = st.nextToken();
    if(t != null) {
      fractionPart = t;
      if(t.length() > n) fractionPart = t.substring(0,n);
    }
    if(st.hasMoreTokens()) exponent = st.nextToken();
    if(exponent.length() > 0) exponent = "E" + exponent;
    return integerPart + dot + fractionPart + exponent;
  }
 
  public String toString() {
    String tabChar = "\t";
    String s = "";
    s = "\nString output for the PeriodicWavelet2DNew Average Object\n\n";
    s = s + "  |  ";
    if(value[0] != null) {
      s = s + tabChar + "axx= " + digitFilter(value[0], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[1] != null) {
      s = s + tabChar + "axy= " + digitFilter(value[1], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[2] != null) {
      s = s + tabChar + "axz= " + digitFilter(value[2], 5);
    } else {
      s = s + tabChar + "null";
    }
    s = s + "  |\n";
    s = s + "  |  ";
    if(value[3] != null) {
      s = s + tabChar + "ayx= " + digitFilter(value[3], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[4] != null) {
      s = s + tabChar + "ayy= " + digitFilter(value[4], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[5] != null) {
      s = s + tabChar + "ayz= " + digitFilter(value[5], 5);
    } else {
      s = s + tabChar + "null";
    }
    s = s + "  |\n";
    s = s + "  |  ";
    if(value[6] != null) {
      s = s + tabChar + "azx= " + digitFilter(value[6], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[7] != null) {
      s = s + tabChar + "azy= " + digitFilter(value[7], 5);
    } else {
      s = s + tabChar + "null";
    }
    if(value[8] != null) {
      s = s + tabChar + "azz= " + digitFilter(value[8], 5);
    } else {
      s = s + tabChar + "null";
    }
    s = s + "  |\n";
    return s;
  }
 
  private String tensorString(double [][] tens) {
    String tabChar = "\t";
    String s = "";
    s = "\nString version of a tensor in an Average Object\n\n";
    s = s + "  |  ";
      s = s + tabChar + "axx= " + digitFilter(tens[0][0], 5);
      s = s + tabChar + "axy= " + digitFilter(tens[0][1], 5);
      s = s + tabChar + "axz= " + digitFilter(tens[0][2], 5);
    s = s + "  |\n";
    s = s + "  |  ";
      s = s + tabChar + "ayx= " + digitFilter(tens[1][0], 5);
      s = s + tabChar + "ayy= " + digitFilter(tens[1][1], 5);
      s = s + tabChar + "ayz= " + digitFilter(tens[1][2], 5);
    s = s + "  |\n";
    s = s + "  |  ";
      s = s + tabChar + "azx= " + digitFilter(tens[2][0], 5);
      s = s + tabChar + "azy= " + digitFilter(tens[2][1], 5);
      s = s + tabChar + "azz= " + digitFilter(tens[2][2], 5);
    s = s + "  |\n";
    return s;
  }

  public void testOutput() {
    System.out.println("isValueOK() test:   " + this.isValueOK());
    System.out.println("getStatus() test:   " + this.getStatus());
    System.out.println("   ");
    System.out.println("double Output: " + this.getdoubleValue());
    System.out.println("isValueOK() test:   " + this.isValueOK());
    System.out.println("getStatus() test:   " + this.getStatus());
    System.out.println("getNumberOfXBlocks() test:   "
                                              + this.getNumberOfXBlocks());
    System.out.println("getNumberOfYBlocks() test:   "
                                              + this.getNumberOfYBlocks());
    System.out.println("getNumberOfZBlocks() test:   "
                                              + this.getNumberOfZBlocks());
    System.out.println("getXLeft() test:             " + this.getXLeft());
    System.out.println("getxLeft() test:             " + this.getxLeft());
    System.out.println("getXRight() test:            " + this.getXRight());
    System.out.println("getxRight() test:            " + this.getxRight());
    System.out.println("getYLeft() test:             " + this.getYLeft());
    System.out.println("getyLeft() test:             " + this.getyLeft());
    System.out.println("getYRight() test:            " + this.getYRight());
    System.out.println("getyRight() test:            " + this.getyRight());
    System.out.println("getZLeft() test:             " + this.getZLeft());
    System.out.println("getzLeft() test:             " + this.getzLeft());
    System.out.println("getZRight() test:            " + this.getZRight());
    System.out.println("getzRight() test:            " + this.getzRight());
    System.out.println("getXCenter() test:           " + this.getXCenter());
    System.out.println("getYCenter() test:           " + this.getYCenter());
    System.out.println("getZCenter() test:           " + this.getZCenter());
    System.out.println("getxCenter() test:           " + this.getxCenter());
    System.out.println("getyCenter() test:           " + this.getyCenter());
    System.out.println("getzCenter() test:           " + this.getzCenter());
    System.out.println("getTotalBlocks() test:       " + this.getTotalBlocks());
    System.out.println("getTag() test:               " + this.getTag());
  }
 
  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *    main method for testing the interface                                *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */
  public static void main(String args[]) {
    int i, j, k, l;

    System.out.println("\n****** Stratified Test #1 Started **************\n");

    PeriodicWaveletAverage2DNew awa;

    // set up the simple 2x2 example
    int nx = 2;
    int ny = 2;
    int nz = 1;
    int ntot = nx * ny * nz;

    // set the data
    Double [][] data = new Double[ntot][9];
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(0.0);
    for(i=0; i<ntot; i++) {
        data[i][0] = new Double(1.0);
        data[i][4] = new Double(1.0);
        data[i][8] = new Double(1.0);
    }
    for(i=0; i<ntot; i=i+2) {
        data[i][0] = new Double(2.0);
        data[i][4] = new Double(2.0);
        data[i][8] = new Double(2.0);
    }

    // the test
    try {
      // set up the object and do the calculations
      awa = new PeriodicWaveletAverage2DNew(data, nx, ny, nz, 1);
      while(!awa.getStatus()) {
        try {Thread.sleep(100);
        } catch(InterruptedException ie) {
          break;
        }
      }
      // print out the results
      System.out.println("Double Output: " + awa.toString());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }

    System.out.println("\n************** End of Test #1 *******************\n");

    System.out.println("\n****** Stratified Test #2 Started **************\n");

    nx = 2;
    ny = 2;
    nz = 1;
    ntot = nx * ny * nz;

    // set the data
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(0.0);
    data[0][0] = new Double(1.0);
    data[0][4] = new Double(1.0);
    data[0][8] = new Double(1.0);
    data[1][0] = new Double(1.0);
    data[1][4] = new Double(1.0);
    data[1][8] = new Double(1.0);
    data[2][0] = new Double(2.0);
    data[2][4] = new Double(2.0);
    data[2][8] = new Double(2.0);
    data[3][0] = new Double(2.0);
    data[3][4] = new Double(2.0);
    data[3][8] = new Double(2.0);

    // the test
    try {
      // set up the object and do the calculations
      awa = new PeriodicWaveletAverage2DNew(data, nx, ny, nz, 1);
      while(!awa.getStatus()) {
        try {Thread.sleep(100);
        } catch(InterruptedException ie) {
          break;
        }
      }
      // print out the results
      System.out.println("Double Output: " + awa.toString());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }

    System.out.println("\n************** End of Test #1 *******************\n");

    System.out.println("\n************* 1-2-3-4 Test Started **************\n");

    nx = 2;
    ny = 2;
    nz = 1;
    ntot = nx * ny * nz;

    // set the data
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(0.0);
    data[0][0] = new Double(1.0);
    data[0][4] = new Double(1.0);
    data[0][8] = new Double(1.0);
    data[1][0] = new Double(2.0);
    data[1][4] = new Double(2.0);
    data[1][8] = new Double(2.0);
    data[2][0] = new Double(3.0);
    data[2][4] = new Double(3.0);
    data[2][8] = new Double(3.0);
    data[3][0] = new Double(4.0);
    data[3][4] = new Double(4.0);
    data[3][8] = new Double(4.0);

    // the test
    try {
      // set up the object and do the calculations
      awa = new PeriodicWaveletAverage2DNew(data, nx, ny, nz, 1);
      while(!awa.getStatus()) {
        try {Thread.sleep(100);
        } catch(InterruptedException ie) {
          break;
        }
      }
      // print out the results
      System.out.println("Double Output: " + awa.toString());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }

    System.out.println("\n************** End 1-2-3-4 Test  ****************\n");

    System.out.println("\n********** 4x4 Stratified Test Started **********\n");

    nx = 4;
    ny = 4;
    nz = 1;
    ntot = nx * ny * nz;

    // set the data
    data = new Double[ntot][9];
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(0.0);
    for(i=0; i<ntot; i++) {
        data[i][0] = new Double(1.0);
        data[i][4] = new Double(1.0);
        data[i][8] = new Double(1.0);
    }

    // the test
    try {
      // set up the object and do the calculations
      awa = new PeriodicWaveletAverage2DNew(data, nx, ny, nz, 1);
      while(!awa.getStatus()) {
        try {Thread.sleep(100);
        } catch(InterruptedException ie) {
          break;
        }
      }
      // print out the results
      System.out.println("Double Output: " + awa.toString());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }

    System.out.println("\n********** End 4x4 Stratified Test  *************\n");

    System.out.println("\n********** 16x16 Random Test Started **********\n");

    nx = 16;
    ny = 16;
    nz = 1;
    ntot = nx * ny * nz;

    // set the data
    data = new Double[ntot][9];
    for(i=0; i<ntot; i++) for(l=0; l<9; l++) data[i][l] = new Double(0.0);
    for(i=0; i<ntot; i++) {
        data[i][0] = new Double(10.0*Math.random());
        data[i][4] = new Double(10.0*Math.random());
        data[i][8] = new Double(10.0*Math.random());
    }

    // the test
    try {
      // set up the object and do the calculations
      awa = new PeriodicWaveletAverage2DNew(data, nx, ny, nz, 1);
      while(!awa.getStatus()) {
        try {Thread.sleep(100);
        } catch(InterruptedException ie) {
          break;
        }
      }
      // print out the results
      System.out.println("Double Output: " + awa.toString());
    } catch(BadArraySizeException base) {
      System.out.println("Bad Array Size Exception");
    }

    System.out.println("\n********** End 16x16 Random Test  *************\n");

  }

  // local variables
  private Double [] value;
  private Double [][] data;
  private int nxhet;
  private int nyhet;
  private int nzhet;
  private int nxnyhet;
  private int nnhet;
  private int powOfTwo = 0;
  private Double xleft;
  private Double xright;
  private Double yleft;
  private Double yright;
  private Double zleft;
  private Double zright;
  private Double [] xgrid;
  private Double [] ygrid;
  private Double [] zgrid;
  private Double xcenter;
  private Double ycenter;
  private Double zcenter;

  // a thread variable for running the neccesary averaging
  private Thread computeThread;
  private boolean done = false;
  private boolean ok = false;

  // the following integer is a tag to identify the object for use by the
  // object which instantitates one of these averaging things
  private int tag;

}
