import java.awt.*;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;
import java.io.*;
public class UpwindSaturationObject extends Object {

  /*
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   *                                                                         *
   *  Constructors for the class                                             *
   *                                                                         *
   * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   */

  // full constructor
  public UpwindSaturationObject(
           double [] u0,             // initial condition
           PointSources ps,          // point sources in the domain
           int nxi,                  // number of nodes in 1st component
           int nyi,                  // number of nodes in 2nd component
           int nzi,                  // number of nodes in 3rd component
           double [] xg,             // explicit grid line values in x
           double [] yg,             // explicit grid line values in y
           double [] zg,             // explicit grid line values in z
           double [] vx,             // explicit grid line values in x
           double [] vy,             // explicit grid line values in y
           double [] vz,             // explicit grid line values in z
           double dt,                // time step to use
           TextArea ta)              // text area for outputing messages
                                 throws WellPosedProblemException {
    // outputArea
    outputArea = ta;
    outputMessage("\nBeginning MFE Solution of the Elliptic Problem.....\n",
                  false);

    outputMessage("\nStarting UpwindSaturationObject\n", false);

    // test the input
    String s;
    if(u0 == null) {
      s = "Null initial condition on the unknown";
      throw new WellPosedProblemException(s);
    }
    pointSources = ps;
    if(pointSources == null) pointSources = new PointSources();
    nx = nxi;
    ny = nyi;
    nz = nzi;
    nxny = nx * ny;
    nnp = nxny * nz;
    if(u0.length != nnp) {
      s = "Initial condition on the unknown has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    uold = new double[nnp];
    unew = new double[nnp];
    for(int i=0; i<nnp; i++) uold[i] = u0[i];

    nxm1 = nx - 1; 
    nym1 = ny - 1; 
    nzm1 = nz - 1; 
    nxp1 = nx + 1;
    nyp1 = ny + 1;
    nzp1 = nz + 1;

    if(xg == null) {
      s = "Null x-grid";
      throw new WellPosedProblemException(s);
    }
    if(xg.length != nxp1) {
      s = "x-grid has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    xgr = new double[nxp1];
    for(int i=0; i<nxp1; i++) xgr[i] = xg[i];

    if(yg == null) {
      s = "Null y-grid";
      throw new WellPosedProblemException(s);
    }
    if(yg.length != nyp1) {
      s = "y-grid has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    ygr = new double[nxp1];
    for(int j=0; j<nyp1; j++) ygr[j] = yg[j];

    if(zg == null) {
      s = "Null z-grid";
      throw new WellPosedProblemException(s);
    }
    if(zg.length != nzp1) {
      s = "z-grid has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    zgr = new double[nxp1];
    for(int j=0; j<nzp1; j++) zgr[j] = zg[j];

    nvelx = nxp1 * ny * nz;
    nvely = nx * nyp1 * nz;
    nvelz = nx * ny * nzp1;

    if(vx == null) {
      s = "Null x-velocity";
      throw new WellPosedProblemException(s);
    }
    if(vx.length != nvelx) {
      s = "x-velocity has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    velx = new double[nvelx];
    for(int i=0; i<nvelx; i++) velx[i] = vx[i];

    if(vy == null) {
      s = "Null y-velocity";
      throw new WellPosedProblemException(s);
    }
    if(vy.length != nvely) {
      s = "y-velocity has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    vely = new double[nvely];
    for(int j=0; j<nvely; j++) vely[j] = vy[j];

    if(vz == null) {
      s = "Null z-velocity";
      throw new WellPosedProblemException(s);
    }
    if(vz.length != nvelz) {
      s = "z-velocity has incompatible dimension";
      throw new WellPosedProblemException(s);
    }
    velz = new double[nvelz];
    for(int k=0; k<nvelz; k++) velz[k] = vz[k];

    outputMessage("Computing the grid for the approximation");
    computeGrid();

    outputMessage("Computing the time step for approximation");
    if(dt <= 0.0) {
      s = "bad time step";
      throw new WellPosedProblemException(s);
    }
    tstep = dt;
    resetTimeStep(true);

    ugridx = new double [nvelx];
    ugridy = new double [nvely];
    ugridz = new double [nvelz];

  }

  private void setOutputArea(TextArea ta) {
    outputArea = ta;
  }

  private void outputMessage(String s) {
    outputMessage(s,true);
  }

  private void outputMessage(String s, boolean append) {
    if(outputArea == null) return;
    if(append) {
      outputArea.append(s);
    } else {
      outputArea.setText(s);
    }
    outputArea.append("\n");
    outputArea.setCaretPosition(10000000);
  }

  public void resetTimeStep(boolean computeDt) {
    if(computeDt) {
      double tmp = computeTimeStep();
      if(tstep > tmp) tstep = tmp;
    }
    if(useFixedTimeStep) tstep = fixedTimeStep;
  }

  public double computeTimeStep() {
    // compute the maximum velocity
    double vmax = 0.0;
    for(int i=0; i<nvelx; i++) {
      double val = Math.abs(velx[i]);
      if(val > vmax) vmax = val;
    }
    for(int j=0; j<nvely; j++) {
      double val = Math.abs(vely[j]);
      if(val > vmax) vmax = val;
    }
    for(int k=0; k<nvelz; k++) {
      double val = Math.abs(velz[k]);
      if(val > vmax) vmax = val;
    }
    // compute the minimum grid size
    double dmin = dx[0];
    for(int i=0; i<nx; i++) {
      double val = Math.abs(dx[i]);
      if(val < dmin) dmin = val;
    }
    for(int j=0; j<ny; j++) {
      double val = Math.abs(dy[j]);
      if(val < dmin) dmin = val;
    }
    for(int k=0; k<nz; k++) {
      double val = Math.abs(dz[k]);
      if(val < dmin) dmin = val;
    }
    // compute a factor that combines the dimensions into the
    // calculation
    double factor = 2.0;
    if(nx > 1) factor = 0.5 * factor;
    if(ny > 1) factor = 0.5 * factor;
    if(nz > 1) factor = 0.5 * factor;
    double tmp = factor * dmin / vmax;
    // return the time step
    return tmp;
  }

  public void setFixedTimeStep(double dt) {
    useFixedTimeStep = true;
    if(dt > 0.0) fixedTimeStep = dt;
  }

  public void setPointSource(PointSources ps) {
    pointSources = ps;
  }

  public void addPointSource(double [] ps) {
    if(ps == null) return;
    if(ps.length != 4) return;
    pointSources.addSource(ps[0], ps[1], ps[2], ps[3]);
  }

  public void addPointSource(double ps0, double ps1, double ps2, double ps3) {
    double [] ps = new double[4];
    ps[0] = ps0;
    ps[1] = ps1;
    ps[2] = ps2;
    ps[3] = ps3;
    addPointSource(ps);
  }

  public void computeGrid() {
    dx = new double[nx];
    for(int i=0; i<nx; i++) dx[i] = xgr[i+1] - xgr[i];
    dy = new double[ny];
    for(int j=0; j<ny; j++) dy[j] = ygr[j+1] - ygr[j];
    dz = new double[nz];
    for(int k=0; k<nz; k++) dz[k] = zgr[k+1] - zgr[k];
    x = new double[nx];
    for(int i=0; i<nx; i++) x[i] = 0.5 * ( xgr[i+1] + xgr[i] );
    y = new double[ny];
    for(int j=0; j<ny; j++) y[j] = 0.5 * ( ygr[j+1] + ygr[j] );
    z = new double[nz];
    for(int k=0; k<nz; k++) z[k] = 0.5 * ( zgr[k+1] + zgr[k] );
  }

  public boolean takeAStep(boolean computeDt) {
    outputMessage("Taking a Step in UpwindSaturationObject", false);
    if(stepTaken) {
      for(int i=0; i<nnp; i++) uold[i] = unew[i];
      stepTaken = false;
    }
    outputMessage("Computing Grid Line Flux in UpwindSaturationObject");
    computeGridFlux();
    outputMessage("Computing Update in UpwindSaturationObject");
    update(computeDt);
    return true;
  }

  public void computeGridFlux() {
    // x component
    for(int k=0; k<nz; k++) {
      int k1 = k * nxp1 * ny;
      int k2 = k * nxny;
      for(int j=0; j<ny; j++) {
        int j1 = j * nxp1 + k1;
        int j2 = j * nx + k2;
        ugridx[j1] = uold[j2] * velx[j1];
        for(int i=1; i<nx; i++) {
          int i1 = i + j1;
          int i2 = i + j2;
          if(velx[i1] >= 0.0) {
            ugridx[i1] = uold[i2-1] * velx[i1];
          } else {
            ugridx[i1] = uold[i2] * velx[i1];
          }
        }
        ugridx[j1+nx] = uold[j2+nx-1] * velx[j1+nx];
      }
    }
    // y component
    for(int i=0; i<nx; i++) {
      int i1 = i * nyp1 * nz;
      int i2 = i;
      for(int k=0; k<nz; k++) {
        int k1 = k * nyp1 + i1;
        int k2 = k * nxny + i2;
        ugridy[k1] = uold[k2] * vely[k1];
        for(int j=1; j<ny; j++) {
          int j1 = j + k1;
          int j2 = j * nx + k2;
          if(vely[j1] >= 0.0) {
            ugridy[j1] = uold[j2-nx] * vely[j1];
          } else {
            ugridy[j1] = uold[j2] * vely[j1];
          }
        }
        ugridy[k1+ny] = uold[k2+nym1*nx] * vely[k1+ny];
      }
    }
    // z component
    for(int j=0; j<ny; j++) {
      int j1 = j * nx * nzp1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int i1 = i * nzp1 + j1;
        int i2 = i + j2;
        ugridz[i1] = uold[i2] * velz[i1];
        for(int k=1; k<nz; k++) {
          int k1 = k + i1;
          int k2 = k * nxny + i2;
          if(velz[k1] >= 0.0) {
            ugridz[k1] = uold[k2-nxny] * velz[k1];
          } else {
            ugridz[k1] = uold[k2] * velz[k1];
          }
        }
        ugridz[i1+nz] = uold[i2+nzm1*nxny] * velz[i1+nz];
      }
    }
  }

  public void update(boolean computeDt) {
    // reset the time step if requested
    resetTimeStep(computeDt);
    // check to see if the step has already been taken
    if(stepTaken) return;
    // x component
    for(int k=0; k<nz; k++) {
      int k1 = k * nxp1 * ny;
      int k2 = k * nxny;
      for(int j=0; j<ny; j++) {
        int j1 = j * nxp1 + k1;
        int j2 = j * nx + k2;
        for(int i=0; i<nx; i++) {
          int i1 = i + j1;
          int i2 = i + j2;
          unew[i2] = uold[i2] - tstep * ( ugridx[i1+1] - ugridx[i1] ) / dx[i];
        }
      }
    }
    // y component
    for(int i=0; i<nx; i++) {
      int i1 = i * nyp1 * nz;
      int i2 = i;
      for(int k=0; k<nz; k++) {
        int k1 = k * nyp1 + i1;
        int k2 = k * nx + i2;
        for(int j=0; j<ny; j++) {
          int j1 = j + k1;
          int j2 = j * nx + k2;
          unew[j2] = unew[j2] - tstep * ( ugridy[j1+1] - ugridy[j1] ) / dy[j];
        }
      }
    }
    // z component
    for(int j=0; j<ny; j++) {
      int j1 = j * nx * nzp1;
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        int i1 = i * nzp1 + j1;
        int i2 = i + j2;
        for(int k=1; k<nz; k++) {
          int k1 = k + i1;
          int k2 = k * nxny + i2;
          unew[k2] = unew[k2] - tstep * ( ugridz[k1+1] - ugridz[k1] ) / dz[k];
        }
      }
    }

    // loop over the point sources to make sure the value is fixed
    int ns = pointSources.getNumberOfSources();
    for(int l=0; l<ns; l++) { 
      double [] ps = (double []) pointSources.getSource(l);
      for(int k=0; k<nz; k++) {

        if(ps[2] >= zgr[k] && ps[2] <= zgr[k+1]) {
          int k2 = k * nxny;
          for(int j=0; j<ny; j++) {
            if(ps[1] >= ygr[j] && ps[1] <= ygr[j+1]) {
              int j2 = j * nx + k2;
              for(int i=0; i<nx; i++) {
                if(ps[0] >= xgr[i] && ps[0] <= xgr[i+1]) {
                  int index = i + j2;
                  unew[index] = ps[3];
                }
              }
            }

          }
        }

      }
    }
    stepTaken = true;
  }

  public double [] getUnew() {
    double [] tmp = new double[nnp];
    for(int i=0; i<nnp; i++) tmp[i] = unew[i];
    return tmp;
  }

  public Image getSolPPMImageXY(Frame f, int kval) {
    if(kval < 0 || kval >= nz) return null;
    return getSolPPMImageXY(f, z[kval]);
  }


  public Image getSolPPMImageXY(Frame f, double zval) {
    outputMessage("Building Image in UpwindSaturationObject", false);
    if(unew == null) return null;
    if(zgr == null) return null;
    int k = 0;
    if(zval < zgr[0]) k = 0;
    for(int i=0; i<nz; i++) {
      if(zval >= zgr[i] && zval < zgr[i+1]) k = i;
    }
    if(zval >= zgr[nz]) k = nz - 1;
    double [] data = new double[nxny];
    int k2 = k * nxny;
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        data[i+j2] = unew[i+j2+k2];
      }
    }
    PPMGraphics ppmGraphics = new PPMGraphics(x, y, data,
                                              PPMGraphics.COLOR_MAP);
    ppmGraphics.setLinearFilter();
    ppmGraphics.setBorder(true);
    ppmGraphics.setBorderValue(10);
    uImage = ppmGraphics.buildImage(f, imageWidth);
    return uImage;
  }

  public boolean writeSolPPMImageXY(Frame f, int kval, File file) {
    if(kval < 0 || kval >= nz) return false;
    return writeSolPPMImageXY(f, z[kval], file);
  }

  public boolean writeSolPPMImageXY(Frame f, double zval, File file) {
    outputMessage("Writing Image in UpwindSaturationObject", false);
    if(unew == null) return false;
    if(zgr == null) return false;
    int k = 0;
    if(zval < zgr[0]) k = 0;
    for(int i=0; i<nz; i++) {
      if(zval >= zgr[i] && zval < zgr[i+1]) k = i;
    }
    if(zval >= zgr[nz]) k = nz - 1;
    double [] data = new double[nxny];
    int k2 = k * nxny;
    for(int j=0; j<ny; j++) {
      int j2 = j * nx;
      for(int i=0; i<nx; i++) {
        data[i+j2] = unew[i+j2+k2];
      }
    }
    PPMGraphics ppmGraphics = new PPMGraphics(x, y, data,
                                              PPMGraphics.COLOR_MAP);
    ppmGraphics.setLinearFilter();
    ppmGraphics.setBorder(true);
    ppmGraphics.setBorderValue(10);
    ppmGraphics.writePPM(file, imageWidth, imageHeight);
    return true;
  }

  private void stopHere() {
    System.out.println("Stop Here");
    try{
      System.in.read();
      System.in.read();
    } catch(IOException ioe) {
    }
  }

  public static void main(String args[]) {

    int nx = 5;
    int ny = 5;
    int nz = 1;
    int nxp1 = nx + 1; 
    int nyp1 = ny + 1; 
    int nzp1 = nz + 1; 
    int nvelx = nxp1 * ny * nz;
    int nvely = nx * nyp1 * nz;
    int nvelz = nx * ny * nzp1;
    int nnp = nx * ny * nz;

    double [] xg = new double [nx+1];
    double dx = 1.0 / ((double) nx);
    xg[0] = 0.0;
    for(int i=1; i<nx+1; i++) {
      xg[i] = xg[i-1] + dx;
    }

    double [] yg = new double [ny+1];
    double dy = 1.0 / ((double) ny);
    yg[0] = 0.0;
    for(int i=1; i<ny+1; i++) {
      yg[i] = yg[i-1] + dy;
    }

    double [] zg = new double [nz+1];
    double dz = 1.0 / ((double) nz);
    zg[0] = 0.0;;
    for(int i=1; i<nz+1; i++) {
      zg[i] = zg[i-1] + dz;
    }

    double [] vx = new double [nvelx];
    for(int i=0; i<nvelx; i++) vx[i] = 1.0;
    double [] vy = new double [nvely];
    for(int j=0; j<nvely; j++) vy[j] = 1.0;
    double [] vz = new double [nvelz];
    for(int k=0; k<nvelz; k++) vz[k] = 0.0;

    double [] u0 = new double[nnp];
    for(int k=0; k<nz; k++) {
      int k2 = k * nx * ny;
      for(int j=0; j<ny; j++) {
        int j2 = j * nx + k2;
        for(int i=0; i<nx; i++) {
          int ind = i + j2;
          if(i<2 && j<2) {
            u0[ind] = 1.0;
          } else {
            u0[ind] = 0.0;
          }
        }
      }
    }

    double dt = 100000.01;

    System.out.println("****************** Upwind Test Case *****************");

    UpwindSaturationObject uso = null;
    try {
      uso =
         new UpwindSaturationObject(u0, null,
                                    nx, ny, nz,
                                    xg, yg, zg,
                                    vx, vy, vz,
                                    dt, null);
    } catch(WellPosedProblemException wppe) {
      System.out.println("Well Posed Problem Exception Thrown");
      System.exit(0);
    }
    for(int i=0; i<100; i++) uso.takeAStep(false);
    double [] u1 = uso.getUnew();
    for(int i=0; i<nnp; i++) {
      System.out.println("input/output:  " + i + "  " + u0[i] + "  " + u1[i]);
    }
    //uso.takeAStep(false);
    //u1 = uso.getUnew();
    //for(int i=0; i<nnp; i++) {
    //  System.out.println("input/output:  " + i + "  " + u0[i] + "  " + u1[i]);
    //}

    //uso.addPointSource(0.5, 0.5, 0.5, 0.75);
    //for(int i=0; i<10; i++) uso.takeAStep(false);
    //u1 = uso.getUnew();
    //for(int i=0; i<nnp; i++) {
    //  System.out.println("input/output:  " + i + "  " + u0[i] + "  " + u1[i]);
    //}

    System.out.println("********************* End Test ********************");

  }

  // some integer parameters
  private int nxm1;
  private int nym1;
  private int nzm1;
  private int nx;
  private int ny;
  private int nz;
  private int nxp1;
  private int nyp1;
  private int nzp1;
  private int nnp;
  private int nxny;
  private int nvelx;
  private int nvely;
  private int nvelz;

  private double [] xgr;
  private double [] ygr;
  private double [] zgr;
  private double [] x;
  private double [] y;
  private double [] z;
  private double [] dx;
  private double [] dy;
  private double [] dz;

  // time dependent piece
  private double tstep = 0.01;
  private boolean useFixedTimeStep = false;
  private double fixedTimeStep;

  // variable stuff
  private double [] uold;
  private double [] unew;
  private PointSources pointSources = null;
  private double [] ugridx;
  private double [] ugridy;
  private double [] ugridz;
  private double [] velx;
  private double [] vely;
  private double [] velz;

  private boolean stepTaken = false;

  private Image uImage = null;
  private int imageWidth = 256;
  private int imageHeight = 256;

  private TextArea outputArea = null;
  public boolean debug = false;

}
